from _decimal import Decimal
from enum import Enum
from typing import Optional, Union, List

from eth_typing import ChecksumAddress

from rubi.contracts.types.events import BaseEvent, EmitOfferEvent, EmitCancelEvent, EmitTakeEvent
from rubi.types.pair import Pair


class OrderSide(Enum):
    """Enumeration representing the order side."""
    BUY = "BUY"
    SELL = "SELL"


class OrderType(Enum):
    """Enumeration representing the order type."""
    MARKET = "MARKET"
    LIMIT = "LIMIT"

    # Only used for events coming from the RubiconMarket
    LIMIT_TAKEN = "LIMIT_TAKEN"
    CANCEL = "CANCEL"


class BaseNewOrder:
    """Base class for representing a new order.

    :param pair_name: The name of the trading pair.
    :type pair_name: str
    :param order_type: The type of the order.
    :type order_type: OrderType
    :param order_side: The side of the order (buy or sell).
    :type order_side: OrderSide
    """

    def __init__(
        self,
        pair_name: str,
        order_type: OrderType,
        order_side: OrderSide,
    ):
        """constructor method."""
        self.pair = pair_name
        self.order_side = order_side
        self.order_type = order_type


class NewMarketOrder(BaseNewOrder):
    """Class representing a new market order.

    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (BUY or SELL).
    :type order_side: OrderSide
    :param size: The size of the order.
    :type size: Decimal
    :param worst_execution_price: The worst execution price for the order (optional). Defaults to 0 if selling and
        10 million if buying as random bounds.
    :type worst_execution_price: Decimal
    """

    def __init__(
        self,
        pair_name: str,
        order_side: OrderSide,
        size: Decimal,
        worst_execution_price: Optional[Decimal],
    ):
        """constructor method."""

        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.MARKET,
            order_side=order_side,

        )
        self.size = size

        if worst_execution_price is None:
            self.worst_execution_price = Decimal("0") if order_side.SELL else Decimal("10") ** Decimal("7")
        else:
            self.worst_execution_price = worst_execution_price


class NewLimitOrder(BaseNewOrder):
    """Class representing a new limit order

    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (buy or sell).
    :type order_side: OrderSide
    :param size: The size of the order.
    :type size: Decimal
    :param price: The price of the order.
    :type price: Decimal
    """

    def __init__(
        self,
        pair_name: str,
        order_side: OrderSide,
        size: Decimal,
        price: Decimal
    ):
        """constructor method."""
        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.LIMIT,
            order_side=order_side,
        )

        self.size = size
        self.price = price


class UpdateLimitOrder(BaseNewOrder):
    """Class representing an update to an existing limit order

    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (BUY or SELL).
    :type order_side: OrderSide
    :param order_id: The ID of the order to update.
    :type order_id: int
    :param size: The updated size of the order.
    :type size: Decimal
    :param price: The updated price of the order.
    :type price: Decimal
    """

    def __init__(
        self,
        pair_name: str,
        order_side: OrderSide,
        order_id: int,
        size: Decimal,
        price: Decimal
    ):
        """constructor method"""
        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.LIMIT,
            order_side=order_side,
        )

        self.order_id = order_id
        self.size = size
        self.price = price


class NewCancelOrder(BaseNewOrder):
    """Class representing a limit order cancellation

    :param pair_name: The name of the trading pair.
    :type pair_name: str
    :param order_side: The side of the order (buy or sell).
    :type order_side: OrderSide
    :param order_id: The ID of the order to cancel.
    :type order_id: int
    """

    def __init__(
        self,
        pair_name: str,
        order_side: OrderSide,
        order_id: int
    ):
        """constructor method"""
        super().__init__(
            pair_name=pair_name,
            order_type=OrderType.LIMIT,
            order_side=order_side,
        )

        self.order_id = order_id


class Transaction:
    """
    Class representing a transaction to be executed on chain

    :param orders: The list of orders to include in the transaction.
    :type orders: List[BaseNewOrder]
    :param nonce: The nonce for the transaction (optional).
    :type nonce: int
    :param gas: The gas limit for the transaction.
    :type gas: int
    :param max_fee_per_gas: The maximum fee per gas for the transaction (optional).
    :type max_fee_per_gas: int
    :param max_priority_fee_per_gas: The maximum priority fee per gas for the transaction (optional).
    :type max_priority_fee_per_gas: int
    """

    def __init__(
        self,
        orders: List[BaseNewOrder],
        nonce: Optional[int] = None,
        gas: int = 350000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ):
        """constructor method"""
        if len(orders) < 1:
            raise Exception("Transaction cannot be instantiated with an empty order list")

        self.orders = orders
        self.nonce = nonce
        self.gas = gas
        self.max_fee_per_gas = max_fee_per_gas
        self.max_priority_fee_per_gas = max_priority_fee_per_gas


class OrderEvent:
    """Class to represent Rubicon Market events as an order

    :param limit_order_id: The ID of the limit order.
    :type limit_order_id: int
    :param limit_order_owner: The owner of the limit order.
    :type limit_order_owner: ChecksumAddress
    :param market_order_owner: The owner of the market order (optional). Only has a value if event is an emitTakeEvent.
    :type market_order_owner: Optional[ChecksumAddress]
    :param pair_name: The name of the pair being traded e.g. WETH/USDC.
    :type pair_name: str
    :param order_side: The side of the order (BUY or SELL).
    :type order_side: OrderSide
    :param order_type: The type of the order (MARKET, LIMIT, LIMIT_TAKEN, or CANCEL).
    :type order_type: OrderType
    :param price: The price of the order.
    :type price: Decimal
    :param size: The size of the order.
    :type size: Decimal
    """

    def __init__(
        self,
        limit_order_id: int,
        limit_order_owner: ChecksumAddress,
        market_order_owner: Optional[ChecksumAddress],
        pair_name: str,
        order_side: OrderSide,
        order_type: OrderType,
        price: Decimal,
        size: Decimal,
    ):
        self.limit_order_id = limit_order_id
        self.limit_order_owner = limit_order_owner
        self.market_order_owner: Optional[ChecksumAddress] = market_order_owner
        self.pair_name = pair_name
        self.order_side = order_side
        self.order_type = order_type
        self.price = price
        self.size = size

    @classmethod
    def from_event(cls, pair: Pair, event: BaseEvent, wallet: ChecksumAddress) -> "OrderEvent":
        """Create an OrderEvent from a BaseEvent emitted by the Rubicon Market.

        :param pair: The asset pair associated with the event.
        :type pair: Pair
        :param event: The event to convert.
        :type event: BaseEvent
        :param wallet: The wallet address associated with the event.
        :type wallet: ChecksumAddress
        :return: The created OrderEvent.
        :rtype: OrderEvent
        :raises Exception: If the event cannot be converted into an OrderEvent. This occurs if the Base event has a type
            other than EmitOfferEvent, EmitCancelEvent or EmitTakeEvent
        """
        if isinstance(event, EmitOfferEvent):
            if pair.bid_identifier == event.pair:
                return cls._build_order(
                    event=event,
                    pair=pair,
                    side=OrderSide.BUY,
                    order_type=OrderType.LIMIT,
                    base_amt=event.buy_amt,
                    quote_amt=event.pay_amt
                )
            else:
                return cls._build_order(
                    event=event,
                    pair=pair,
                    side=OrderSide.SELL,
                    order_type=OrderType.LIMIT,
                    base_amt=event.pay_amt,
                    quote_amt=event.buy_amt
                )
        elif isinstance(event, EmitCancelEvent):
            if pair.bid_identifier == event.pair:
                return cls._build_order(
                    event=event,
                    pair=pair,
                    side=OrderSide.BUY,
                    order_type=OrderType.CANCEL,
                    base_amt=event.buy_amt,
                    quote_amt=event.pay_amt
                )
            else:
                return cls._build_order(
                    event=event,
                    pair=pair,
                    side=OrderSide.SELL,
                    order_type=OrderType.CANCEL,
                    base_amt=event.pay_amt,
                    quote_amt=event.buy_amt
                )
        elif isinstance(event, EmitTakeEvent):
            # This is nuanced as we can either receive a take event for a market order or limit order we placed. When
            # we take the bid_identifier indicates selling into the bids while the ask identifier indicates buying
            # into the asks. While the reverse is true for when we make.
            if pair.bid_identifier == event.pair:
                return cls._build_order(
                    event=event,
                    pair=pair,
                    side=OrderSide.BUY if wallet == event.maker else OrderSide.SELL,
                    order_type=OrderType.LIMIT_TAKEN if wallet == event.maker else OrderType.MARKET,
                    base_amt=event.give_amt,
                    quote_amt=event.take_amt
                )
            else:
                return cls._build_order(
                    event=event,
                    pair=pair,
                    side=OrderSide.SELL if wallet == event.maker else OrderSide.BUY,
                    order_type=OrderType.LIMIT_TAKEN if wallet == event.maker else OrderType.MARKET,
                    base_amt=event.take_amt,
                    quote_amt=event.give_amt
                )
        else:
            Exception(f"{event.__class__} cannot be converted into an OrderEvent")

    @classmethod
    def _build_order(
        cls,
        event: Union[EmitOfferEvent, EmitCancelEvent, EmitTakeEvent],
        pair: Pair,
        side: OrderSide,
        order_type: OrderType,
        base_amt: int,
        quote_amt: int
    ) -> "OrderEvent":
        """Build an OrderEvent from event data.

        :param event: The event data.
        :type event: Union[EmitOfferEvent, EmitCancelEvent, EmitTakeEvent]
        :param pair: The asset pair associated with the event.
        :type pair: Pair
        :param side: The order side.
        :type side: OrderSide
        :param order_type: The order type.
        :type order_type: OrderType
        :param base_amt: The base amount of the order.
        :type base_amt: int
        :param quote_amt: The quote amount of the order.
        :type quote_amt: int
        :return: The constructed OrderEvent.
        :rtype: OrderEvent
        """
        size = pair.base_asset.to_decimal(base_amt)
        price = pair.quote_asset.to_decimal(quote_amt) / size

        return cls(
            limit_order_id=event.id,
            limit_order_owner=event.maker,
            market_order_owner=event.taker if isinstance(event, EmitTakeEvent) else None,
            pair_name=pair.name,
            order_side=side,
            order_type=order_type,
            size=size,
            price=price
        )

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))



from subgrounds import Subgrounds
from rubi.data.sources.aid import AidData, SuperAidData
from rubi.data.sources.market import MarketData
from rubi.data.sources.helper import Gas, Price, networks 
from rubi.data.processing.user import User, SuperUser
from rubi.data.processing.aid import AidProcessing

# TODO this needs significant love.
class Data: 
    """this class acts as the main access point to a variety of data and data tooling for the Rubicon protocol. it acts as a data processing layer built using the subgrounds library and the subgraphs maintained at the follwing repo: https://github.com/RubiconDeFi/rubi-subgraphs
    """

    def __init__(self, super=False): 
        """constructor method. creates a subgrounds object that is then used to initialize a variety of data sources and tooling

        :param super: whether the data object is a super data object, defaults to False
        :type super: bool, optional
        """
        self.subgrounds = Subgrounds()
        self.price = Price()
        self.networks = networks

        # initialize the data sources
        self.market_optimism = MarketData(self.subgrounds, 10)
        self.market_optimism_goerli = MarketData(self.subgrounds, 420)
        
        # initialize the data processing if the data object is not a super data object
        if not super:
            self.user = User(self.subgrounds, self.market_optimism)
            self.market_aid_optimism = AidData(self.subgrounds, 10) 
            self.market_aid_optimism_goerli = AidData(self.subgrounds, 420)

            # TODO: we most likely want to rearrange this so that the processing class uses the AidData class as a parent class
            # this issue here is that we are using a parent/child relationship already to distinguish between a web3 enabled data object and a non web3 enabled data object
            # we should probably simplify this to just use a conditional statement to determine whether or not we are using a web3 enabled data object
            # if anyone has a lot of thoughts here please feel free to open an issue on the repo :)
            self.market_aid_optimism_processing = AidProcessing(self.subgrounds, 10, self.market_aid_optimism)
            self.market_aid_optimism_goerli_processing = AidProcessing(self.subgrounds, 420, self.market_aid_optimism_goerli)

class SuperData(Data):
    """this class acts as an extension of the data class with additional functionality that is enabled by being connected to a node. 
    """

    def __init__(self, w3): 
        """constructor method. creates a subgrounds object that is then used to initialize a variety of data sources and tooling
        """

        # initialize the data class
        super().__init__(super=True)

        # set class variables
        self.w3 = w3
        self.gas = Gas(self.w3)

        # initialize the data processing
        self.user = SuperUser(self.w3, self.subgrounds, self.market_optimism)
        
        # TODO: see if there is a more ellegant want to do this, we are going to split these based upon which network we are on 
        chain = w3.eth.chain_id

        if chain == 10:
            self.market_aid_optimism = SuperAidData(self.w3, self.subgrounds, 10) 
            self.market_aid_optimism_goerli = AidData(self.subgrounds, 420)

            self.market_aid_optimism_processing = AidProcessing(self.subgrounds, 10, self.market_aid_optimism)
            self.market_aid_optimism_goerli_processing = AidProcessing(self.subgrounds, 420, self.market_aid_optimism_goerli)
        elif chain == 420:
            self.market_aid_optimism = AidData(self.subgrounds, 10) 
            self.market_aid_optimism_goerli = SuperAidData(self.w3, self.subgrounds, 420)

            self.market_aid_optimism_processing = AidProcessing(self.subgrounds, 10, self.market_aid_optimism)
            self.market_aid_optimism_goerli_processing = AidProcessing(self.subgrounds, 420, self.market_aid_optimism_goerli)


import json
import os
from _decimal import Decimal
from typing import Optional
from pkg_resources import resource_stream

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import Contract
from web3.types import ABI

from rubi.contracts.base_contract import BaseContract
from rubi.network import Network


class ERC20(BaseContract):
    """this class represents a contract that implements the ERC20 standard. it is used to read the contract instance.
    if a wallet and key are passed in instantiation then this class can also be used to write to the contract instance.

    :param w3: Web3 instance
    :type w3: Web3
    :param contract: Contract instance
    :type contract: Contract
    :param wallet: a wallet address of the signer (optional, default is None)
    :type wallet: Optional[ChecksumAddress]
    :param key: the private key of the signer (optional, default is None)
    :type key: Optional[str]
    """

    def __init__(
        self,
        w3: Web3,
        contract: Contract,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ):
        """constructor method"""
        super().__init__(
            w3=w3,
            contract=contract,
            wallet=wallet,
            key=key
        )

        self.name: str = self.name()
        self.symbol: str = self.symbol()
        self.decimal: int = self.decimals()

    @classmethod
    def from_network(
        cls,
        name: str,
        network: Network,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ) -> "ERC20":
        """Create an ERC20 instance based on a Network instance and token name.

        :param name: The name of the token.
        :type name: str
        :param network: A Network instance.
        :type network: Network
        :param wallet: Optional wallet address to use for interacting with the contract (optional, default is None).
        :type wallet: Optional[ChecksumAddress]
        :param key: Optional private key for the wallet (optional, default is None).
        :type key: Optional[str]
        :return: An ERC20 instance based on the Network instance and token name.
        :rtype: ERC20
        :raises Exception: If the token name does not exist in the network configuration.
        :raises Exception: If the ERC20.json ABI file is not found in the network_config folder.
        """

        if network.token_addresses[name] is None:
            raise Exception(f"{name} in not a valid token according to the network config.")

        abi: ABI

        try:
            with resource_stream('rubi', "../network_config/ERC20.json") as f:
                abi = json.load(f)

        except FileNotFoundError:
            raise Exception("ERC20.json abi not found. this file should in the network_config folder")

        return cls.from_address_and_abi(
            w3=network.w3,
            address=network.token_addresses[name],
            contract_abi=abi,
            wallet=wallet,
            key=key
        )

    ######################################################################
    # read calls
    ######################################################################

    # allowance(owner (address), spender (address)) -> uint256
    def allowance(self, owner: ChecksumAddress, spender: ChecksumAddress) -> int:
        """Reads the allowance of the spender from the owner for the erc20 contract

        :param owner: address that owns the erc20 tokens
        :type owner: str
        :param spender: address that is allowed to spend the erc20 tokens
        :type spender: str
        :return: the allowance of the spender from the owner for the contract, in the integer representation of the
            token
        :rtype: int
        """

        return self.contract.functions.allowance(owner, spender).call()

    # balanceOf(account (address)) -> uint256
    def balance_of(self, account: ChecksumAddress) -> int:
        """Reads the erc20 balance of the account

        :param account: the address of the account to read the balance of
        :type account: str
        :return: the balance of the account, in the integer representation of the token
        :rtype: int
        """

        return self.contract.functions.balanceOf(account).call()

    # totalSupply() -> uint256
    def total_supply(self) -> int:
        """Reads the total supply of the erc20 token

        :return: the total supply of the erc20 token, in the integer representation of the token
        :rtype: int
        """

        return self.contract.functions.totalSupply().call()

    # decimals() -> uint256
    def decimals(self) -> int:
        """Reads the number of decimals of the erc20 token - warning this is not a constant function, so it may result
        in an error in its current implementation

        :return: the number of decimals of the erc20 token
        :rtype: int
        """

        return self.contract.functions.decimals().call()

    # name() -> string
    def name(self) -> str:
        """Reads the name of the erc20 token

        :return: the name of the erc20 token
        :rtype: str
        """

        return self.contract.functions.name().call()

    # symbol() -> string
    def symbol(self) -> str:
        """Reads the symbol of the erc20 token

        :return: the symbol of the erc20 token
        :rtype: str
        """

        return self.contract.functions.symbol().call()

    ######################################################################
    # write calls
    ######################################################################

    # approve(spender (address), amount (uint256)) -> bool
    def approve(
        self,
        spender: ChecksumAddress,
        amount: int,
        nonce: Optional[int] = None,
        gas: int = 100000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Approves the spender to spend the amount of the erc20 token from the signer's wallet

        :param spender: address of the spender
        :type spender: ChecksumAddress
        :param amount: amount of the erc20 token to approve the spender to spend
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        approve = self.contract.functions.approve(spender, amount)

        return self._default_transaction_handler(
            instantiated_contract_function=approve,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # transfer(recipient (address), amount (uint256)) -> bool
    def transfer(
        self,
        recipient: ChecksumAddress,
        amount: int,
        nonce: Optional[int] = None,
        gas: int = 100000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Transfers the amount of the erc20 token to the recipient

        :param recipient: address of the recipient
        :type recipient: ChecksumAddress
        :param amount: amount of the erc20 token to transfer
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        transfer = self.contract.functions.transfer(recipient, amount)

        return self._default_transaction_handler(
            instantiated_contract_function=transfer,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # transferFrom(sender (address), recipient (address), amount (uint256)) -> bool
    def transfer_from(
        self,
        sender: ChecksumAddress,
        recipient: ChecksumAddress,
        amount: int,
        nonce: Optional[int] = None,
        gas: int = 100000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Transfers the amount of the erc20 token from the sender to the recipient

        :param sender: address of the sender
        :type sender: ChecksumAddress
        :param recipient: address of the recipient
        :type recipient: ChecksumAddress
        :param amount: amount of the erc20 token to transfer
        :type amount: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """

        transfer_from = self.contract.functions.transferFrom(sender, recipient, amount)

        return self._default_transaction_handler(
            instantiated_contract_function=transfer_from,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # increaseAllowance(spender (address), addedValue (uint256)) -> bool
    def increase_allowance(
        self,
        spender: ChecksumAddress,
        added_value: int,
        nonce: Optional[int] = None,
        gas: int = 100000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Increases the allowance of the spender by the added_value

        :param spender: address of the spender
        :type spender: ChecksumAddress
        :param added_value: amount to increase the allowance by, in the integer representation of the erc20 token
        :type added_value: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        increase_allowance = self.contract.functions.increaseAllowance(spender, added_value)

        return self._default_transaction_handler(
            instantiated_contract_function=increase_allowance,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    # decreaseAllowance(spender (address), subtractedValue (uint256)) -> bool
    def decrease_allowance(
        self,
        spender: ChecksumAddress,
        subtracted_value: int,
        nonce: Optional[int] = None,
        gas: int = 100000,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Decreases the allowance of the spender by the subtracted_value

        :param spender: address of the spender
        :type spender: ChecksumAddress
        :param subtracted_value: amount to decrease the allowance by, in the integer representation of the erc20 token
        :type subtracted_value: int
        :param nonce: nonce of the transaction, defaults to calling the chain state to get the nonce.
            (optional, default is None)
        :type nonce: Optional[int]
        :param gas: gas limit of the transaction, defaults to a very high estimate made when writing the class
        :type gas: int
        :param max_fee_per_gas: max fee that can be paid for gas, defaults to
            max_priority_fee (from chain) + (2 * base fee per gas of latest block) (optional, default is None)
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: max priority fee that can be paid for gas, defaults to calling the chain to
            estimate the max_priority_fee_per_gas (optional, default is None)
        :type max_priority_fee_per_gas: Optional[int]
        :return: transaction hash
        :rtype: str
        """
        decrease_allowance = self.contract.functions.decreaseAllowance(spender, subtracted_value)

        return self._default_transaction_handler(
            instantiated_contract_function=decrease_allowance,
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

    ######################################################################
    # helper methods
    ######################################################################

    def to_decimal(self, number: int) -> Decimal:
        """Converts an integer representation of the token to a Decimal representation of the token by dividing the
        integer by 10 to the power of the number of decimals of the token

        :param number: the integer representation of the token
        :type number: int
        :return: the Decimal representation of the token
        :rtype: Decimal
        """

        if number == 0:
            return Decimal("0")
        else:
            return Decimal(number) / Decimal(10 ** self.decimal)

    def to_integer(self, number: Decimal) -> int:
        """Converts a Decimal representation of the token to an integer representation of the token by multiplying the
        float by 10 to the power of the number of decimals of the token

        :param number: the Decimal representation of the token
        :type number: Decimal
        :return: the integer representation of the token
        :rtype: int
        """

        if number == Decimal("0"):
            return 0
        else:
            return int(number * (10 ** self.decimal))





class Process:
    """this class is really just meant to serve as a place to aggregate useful functionality for processing data. the goal in 
    formatting helper functions in this way is to keep the main processing class for each data source as clean and interoperable as possible.
    """

    def __init__(self):
        self.cowboy = 'hello, defi cowboy. lets get this bread'

    def get_closest_timestamp_value(self, values, timestamp, granularity=60, n=10): 
        """this function takes a dictionary of values (assumes the keys are timestamps) and a timestamp and returns the value of the dictionary that is closest to the timestamp.

        :param values: a dictionary of values with the keys being timestamps
        :type values: dict
        :param timestamp: the timestamp that is of interest
        :type timestamp: int
        :param granularity: the granularity of the timestamps of interest, defaults to 60
        :type granularity: int, optional
        :param n: the optimized search default based upon granularity of which to search , defaults to 10
        :type n: int, optional
        :return: the value of the dictionary that is closest to the timestamp key pair that matches
        :rtype: value
        """

        # based on the granularity of the time perids, get the starting period value 
        timestamp = (timestamp // granularity) * granularity

        # get the keys of the dictionary 
        keys = list(values.keys())

        try: 
            value = values[timestamp]
            return value
        except:
            pass
        
        attempt = 1
        while attempt < n: 
            try: 
                value = values[timestamp + granularity * attempt]
                return value
            except:
                pass

            try:
                value = values[timestamp - granularity * attempt]
                return value
            except:
                pass

            attempt += 1
        
        # if this still doesn't work, search the entire list of keys
        min = abs(keys[0] - timestamp)
        value = values[keys[0]]
        for key in keys[1:]:
            if abs(key - timestamp) < min:
                min = abs(key - timestamp)
                value = values[key]

        return value


import logging as log
from abc import ABC, abstractmethod
from typing import Dict, Any, Optional, Type, TypeVar, Union

from eth_typing import ChecksumAddress, HexStr
from eth_utils import add_0x_prefix
from web3._utils.filters import LogFilter  # noqa
from web3.contract import Contract
from web3.types import EventData

# To solve for circular dependencies
_RubiconMarket = TypeVar("_RubiconMarket")
_RubiconRouter = TypeVar("_RubiconRouter")


class BaseEvent(ABC):
    """Base class for events to define the structure of an Event from a Rubicon contract.
    """

    def __init__(self, block_number: int, **args):
        """Initialize a BaseEvent instance.

        :param block_number: The block number of the event.
        :type block_number: int
        :param args: Additional arguments for the event.
        :type args: dict
        """
        self.block_number = block_number

    @staticmethod
    @abstractmethod
    def get_event_contract(market: _RubiconMarket, router: _RubiconRouter) -> Union[_RubiconMarket, _RubiconRouter]:
        """Abstract method to determine the contract an event corresponds to. Must be overridden in subclasses.

        :param market: The RubiconMarket instance.
        :type market: _RubiconMarket
        :param router: The RubiconRouter instance.
        :type router: _RubiconRouter
        :return: The contract corresponding to the event.
        :rtype: Union[_RubiconMarket, _RubiconRouter]
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def create_event_filter(contract: Contract, argument_filters: Optional[Dict[str, Any]] = None) -> LogFilter:
        """Abstract method to create an event filter for the given contract with optional argument filters. Must be
        overridden in each event subclass.

        :param contract: The contract instance.
        :type contract: Contract
        :param argument_filters: Optional filters. Only events that match these filters will be returned when the filter
            is queried.
        :type argument_filters: Optional[Dict[str, Any]]
        :return: The created event filter.
        :rtype: LogFilter
        """
        raise NotImplementedError()

    @classmethod
    def default_handler(cls, pair_name: str, event_type: Type["BaseEvent"], event_data: EventData) -> None:
        """A default event handler. If no handler is provided then this one is used.

        :param pair_name: The name of the pair.
        :type pair_name: str
        :param event_type: The type of the event.
        :type event_type: Type["BaseEvent"]
        :param event_data: The data of the event.
        :type event_data: EventData
        """
        log.info(event_data)

    @staticmethod
    @abstractmethod
    def default_filters(bid_identifier: str, ask_identifier: str, wallet: ChecksumAddress) -> dict:
        """Get the default filters for an event. These are used if no argument filters are provided. By default, these
        filters make sure we only receive events that relate to us. E.g on markets we care about.

        :param bid_identifier: The identifier for bid events.
        :type bid_identifier: str
        :param ask_identifier: The identifier for ask events.
        :type ask_identifier: str
        :param wallet: The wallet address.
        :type wallet: ChecksumAddress
        :return: The default filters.
        :rtype: dict
        """
        raise NotImplementedError()

    # noinspection PyMethodMayBeStatic
    def client_filter(self, wallet: ChecksumAddress) -> bool:
        """Filter function for client-side filtering. By default, no filtering is done but this can optionally be
        overwritten by each subclass.

        :param wallet: The wallet address.
        :type wallet: ChecksumAddress
        :return: True if the event passes the filter, False otherwise.
        :rtype: bool
        """
        return True

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


######################################################################
# market events
######################################################################

class BaseMarketEvent(BaseEvent, ABC):
    """This class is a base class for all MarketEvents"""

    def __init__(self, id: bytes, pair: bytes, **args):
        """Initialize a BaseMarketEvent instance.

        :param id: The event ID.
        :type id: bytes
        :param pair: The event bid or ask pair identifier.
        :type pair: bytes
        :param args: Additional arguments for the event.
        :type args: dict
        """
        super().__init__(**args)
        self.id = int(id.hex(), 16)
        self.pair = add_0x_prefix(HexStr(pair.hex()))

    @staticmethod
    def get_event_contract(market: _RubiconMarket, router: _RubiconRouter) -> Union[_RubiconMarket, _RubiconRouter]:
        """implementation of BaseEvent get_event_contract"""
        return market


class EmitOfferEvent(BaseMarketEvent):
    """Event emitted whenever a new offer is made on the RubiconMarket
    """

    def __init__(
        self,
        maker: ChecksumAddress,
        pay_gem: ChecksumAddress,
        buy_gem: ChecksumAddress,
        pay_amt: int,
        buy_amt: int,
        **args
    ):
        """Initialize an EmitOfferEvent instance.

       :param maker: The maker address.
       :type maker: ChecksumAddress
       :param pay_gem: The address of the token to be paid.
       :type pay_gem: ChecksumAddress
       :param buy_gem: The address of the token to be bought.
       :type buy_gem: ChecksumAddress
       :param pay_amt: The amount to be paid.
       :type pay_amt: int
       :param buy_amt: The amount to be bought.
       :type buy_amt: int
       :param args: Additional arguments for the event.
       :type args: dict
       """
        super().__init__(**args)
        self.maker = maker
        self.pay_gem = pay_gem
        self.buy_gem = buy_gem
        self.pay_amt = pay_amt
        self.buy_amt = buy_amt

    @staticmethod
    def create_event_filter(contract: Contract, argument_filters: Optional[Dict[str, Any]] = None) -> LogFilter:
        """implementation of BaseEvent create_event_filter"""
        return contract.events.emitOffer.create_filter(argument_filters=argument_filters, fromBlock="latest")

    @staticmethod
    def default_filters(bid_identifier: str, ask_identifier: str, wallet: ChecksumAddress) -> dict:
        """implementation of BaseEvent default_filters"""
        filters = {"pair": [bid_identifier, ask_identifier], "maker": wallet}

        return {key: value for key, value in filters.items() if value is not None}


class EmitTakeEvent(BaseMarketEvent):
    """Event emitted whenever an offer is taken by a market order on the RubiconMarket
    """

    def __init__(
        self,
        maker: ChecksumAddress,
        taker: ChecksumAddress,
        pay_gem: ChecksumAddress,
        buy_gem: ChecksumAddress,
        take_amt: int,
        give_amt: int,
        **args
    ):
        """Initialize an EmitTakeEvent instance.

        :param maker: The maker address.
        :type maker: ChecksumAddress
        :param taker: The taker address.
        :type taker: ChecksumAddress
        :param pay_gem: The address of the token that was paid.
        :type pay_gem: ChecksumAddress
        :param buy_gem: The address of the token that was bought.
        :type buy_gem: ChecksumAddress
        :param take_amt: The amount that was taken.
        :type take_amt: int
        :param give_amt: The amount that was given.
        :type give_amt: int
        :param args: Additional arguments for the event.
        :type args: dict
        """
        super().__init__(**args)

        self.maker = maker
        self.taker = taker
        self.pay_gem = pay_gem
        self.buy_gem = buy_gem
        self.take_amt = take_amt
        self.give_amt = give_amt

    @staticmethod
    def create_event_filter(contract: Contract, argument_filters: Optional[Dict[str, Any]] = None) -> LogFilter:
        """implementation of BaseEvent create_event_filter"""
        return contract.events.emitTake.create_filter(argument_filters=argument_filters, fromBlock="latest")

    @staticmethod
    def default_filters(bid_identifier: HexStr, ask_identifier: HexStr, wallet: ChecksumAddress) -> dict:
        """implementation of BaseEvent default_filters"""
        filters = {"pair": [bid_identifier, ask_identifier]}

        return {key: value for key, value in filters.items() if value is not None}

    def client_filter(self, wallet: ChecksumAddress) -> bool:
        """overwriting of BaseEvent client_filter to only filter when our wallet is either the maker or taker"""
        return wallet is None or (self.maker == wallet or self.taker == wallet)


class EmitCancelEvent(BaseMarketEvent):
    """Event emitted whenever an offer is cancelled on the RubiconMarket
    """

    def __init__(
        self,
        maker: ChecksumAddress,
        pay_gem: ChecksumAddress,
        buy_gem: ChecksumAddress,
        pay_amt: int,
        buy_amt: int,
        **args
    ):
        """Initialize an EmitCancelEvent instance.

        :param maker: The maker address.
        :type maker: ChecksumAddress
        :param pay_gem: The address of the token to be paid of the cancelled offer.
        :type pay_gem: ChecksumAddress
        :param buy_gem: The address of the token to be bought of the cancelled offer.
        :type buy_gem: ChecksumAddress
        :param pay_amt: The amount to be paid of the cancelled offer.
        :type pay_amt: int
        :param buy_amt: The amount to be bought of the cancelled offer.
        :type buy_amt: int
        :param args: Additional arguments for the event.
        :type args: dict
        """
        super().__init__(**args)

        self.maker = maker
        self.pay_gem = pay_gem
        self.buy_gem = buy_gem
        self.pay_amt = pay_amt
        self.buy_amt = buy_amt

    @staticmethod
    def create_event_filter(contract: Contract, argument_filters: Optional[Dict[str, Any]] = None) -> LogFilter:
        """implementation of BaseEvent create_event_filter"""
        return contract.events.emitCancel.create_filter(argument_filters=argument_filters, fromBlock="latest")

    @staticmethod
    def default_filters(bid_identifier: str, ask_identifier: str, wallet: ChecksumAddress) -> dict:
        """implementation of BaseEvent default_filters"""
        filters = {"pair": [bid_identifier, ask_identifier], "maker": wallet}

        return {key: value for key, value in filters.items() if value is not None}


class EmitFeeEvent(BaseMarketEvent):
    """Event emitted whenever an offer is taken on the RubiconMarket that results in a fee being paid to the maker.
    """

    def __init__(
        self,
        taker: ChecksumAddress,
        feeTo: ChecksumAddress,
        asset: ChecksumAddress,
        feeAmt: int,
        **args
    ):
        """Initialize an EmitFeeEvent instance.

        :param taker: The taker address.
        :type taker: ChecksumAddress
        :param feeTo: The address to receive the fee.
        :type feeTo: ChecksumAddress
        :param asset: The address of the asset for the fee.
        :type asset: ChecksumAddress
        :param feeAmt: The amount of the fee.
        :type feeAmt: int
        :param args: Additional arguments for the event.
        :type args: dict
        """
        super().__init__(**args)

        self.taker = taker
        self.fee_to = feeTo
        self.asset = asset
        self.fee_amt = feeAmt

    @staticmethod
    def create_event_filter(contract: Contract, argument_filters: Optional[Dict[str, Any]] = None) -> LogFilter:
        """implementation of BaseEvent create_event_filter"""
        return contract.events.emitFee.create_filter(argument_filters=argument_filters, fromBlock="latest")

    @staticmethod
    def default_filters(bid_identifier: str, ask_identifier: str, wallet: ChecksumAddress) -> dict:
        """implementation of BaseEvent default_filters"""
        filters = {"pair": [bid_identifier, ask_identifier], "fee_to": wallet}

        return {key: value for key, value in filters.items() if value is not None}


class EmitDeleteEvent(BaseMarketEvent):
    """Event emitted whenever an offer is fully taken by a market order on the RubiconMarket resulting in the offer
    being closed
    """

    def __init__(self, maker: ChecksumAddress, **args):
        """Initialize an EmitDeleteEvent instance.

        :param maker: The maker address.
        :type maker: ChecksumAddress
        :param args: Additional arguments for the event.
        :type args: dict
        """
        super().__init__(**args)

        self.maker = maker

    @staticmethod
    def create_event_filter(contract: Contract, argument_filters: Optional[Dict[str, Any]] = None) -> LogFilter:
        """implementation of BaseEvent create_event_filter"""
        return contract.events.emitDelete.create_filter(argument_filters=argument_filters, fromBlock="latest")

    @staticmethod
    def default_filters(bid_identifier: str, ask_identifier: str, wallet: ChecksumAddress) -> dict:
        """implementation of BaseEvent default_filters"""
        filters = {"pair": [bid_identifier, ask_identifier], "maker": wallet}

        return {key: value for key, value in filters.items() if value is not None}


######################################################################
# router events
######################################################################

class EmitSwap(BaseEvent):
    """Event emitted whenever swap is executed on the RubiconRouter
    """

    def __init__(
        self,
        recipient: ChecksumAddress,
        inputERC20: ChecksumAddress,
        targetERC20: ChecksumAddress,
        pair: bytes,
        inputAmount: int,
        realizedFill: int,
        hurdleBuyAmtMin: int,
        **args
    ):
        """Initialize an EmitSwap instance.

        :param recipient: The address of the recipient of the swap.
        :type recipient: ChecksumAddress
        :param inputERC20: The address of the input ERC20 token.
        :type inputERC20: ChecksumAddress
        :param targetERC20: The address of the target ERC20 token.
        :type targetERC20: ChecksumAddress
        :param pair: The bid/offer identifier of the pair of tokens being swapped (first_token/last_token).
        :type pair: bytes
        :param inputAmount: The amount of input tokens being swapped.
        :type inputAmount: int
        :param realizedFill: The realized fill amount.
        :type realizedFill: int
        :param hurdleBuyAmtMin: The minimum hurdle buy amount.
        :type hurdleBuyAmtMin: int
        :param args: Additional arguments for the event.
        :type args: dict
        """
        super().__init__(**args)

        self.recipient = recipient
        self.inputERC20 = inputERC20
        self.targetERC20 = targetERC20
        self.pair = pair
        self.inputAmount = inputAmount
        self.realizedFill = realizedFill
        self.hurdleBuyAmtMin = hurdleBuyAmtMin

    @staticmethod
    def get_event_contract(market: _RubiconMarket, router: _RubiconRouter) -> Union[_RubiconMarket, _RubiconRouter]:
        """implementation of BaseEvent get_event_contract"""
        return router

    @staticmethod
    def create_event_filter(contract: Contract, argument_filters: Optional[Dict[str, Any]] = None) -> LogFilter:
        """implementation of BaseEvent create_event_filter"""
        return contract.events.emitSwap.create_filter(argument_filters=argument_filters, fromBlock="latest")

    @staticmethod
    def default_filters(bid_identifier: str, ask_identifier: str, wallet: ChecksumAddress) -> dict:
        """implementation of BaseEvent default_filters"""
        filters = {"recipient": wallet}

        return {key: value for key, value in filters.items() if value is not None}



import os
import json

class OptimismMainnet:
    """this class represents a rolodex of relevant informatino regarding the Optimism Mainnet network, such as contract addresses, token addresses, and abis."""

    # TODO: set these to be dynamically pulled in a trustless fashion
    def __init__(self):
        self.network = 'Optimism'
        self.chain_id = 10
        self.currency = 'ETH'
        self.rpc_url = 'https://mainnet.optimism.io'
        self.explorer_url = 'https://optimistic.etherscan.io'
        
        # set the rubicon contract addresses
        self.market = '0x7a512d3609211e719737E82c7bb7271eC05Da70d'
        self.router = '0x7Af14ADc8Aea70f063c7eA3B2C1AD0D7A59C4bFf'
        self.pair = '0xF8780E00Ce8ed2e79aeC10908a169900eD1D4AFe'
        self.factory = '0x267D94C6e67e4436EFfE092b08d040cFF36B2DA7'

        # set up some common asset addresses that may be used 
        # TODO: it would be nice to be able to dynamically pull in a token list in a trustless fashion and populate this list
        self.weth = '0x4200000000000000000000000000000000000006'
        self.wbtc = '0x68f180fcCe6836688e9084f035309E29Bf0A2095'
        self.usdc = '0x7F5c764cBc14f9669B88837ca1490cCa17c31607'
        self.dai = '0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1'
        self.usdt = '0x94b008aA00579c1307B0EF2c499aD98a8ce58e58'
        self.snx = '0x8700dAec35aF8Ff88c16BdF0418774CB3D7599B4'
        self.op = '0x4200000000000000000000000000000000000042'

        # set the subgraph query endpoints
        # TODO: this will been to be updated in the future to deal with GRT payments for subgraph queries
        # it would be nice if through this SDK we managed a synergistic relationship between queriers, for common queries that are supported in the SDK 
        # we could possibly hash the most recent query response to an IPFS instance that is then udpated by subsequent qureries, negating the need to query the entire history 
        # of the subgraph, and only querying, and paying, for the most recent data. 
        self.rubicon_market_light = 'https://api.thegraph.com/subgraphs/name/denverbaumgartner/rubiconmarket-light-optimism'
        # self.rubicon_metrics = 'https://api.thegraph.com/subgraphs/name/denverbaumgartner/rubiconmetricsoptimism'
        self.boiler_plate = 'https://api.thegraph.com/subgraphs/name/denverbaumgartner/boilerplateoptimism'
        self.market_aid = 'https://api.thegraph.com/subgraphs/name/denverbaumgartner/optimismmarketaid'

        # create a dictionary of token addresses and symbols to be mapped to a coinbase usd base ticker
        self.coinbase_tickers = {
            '0x4200000000000000000000000000000000000006': 'ETH-USD',
            '0x68f180fcCe6836688e9084f035309E29Bf0A2095': 'BTC-USD',
            '0x7F5c764cBc14f9669B88837ca1490cCa17c31607': 'USDC-USD',
            '0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1': 'DAI-USD',
            '0x94b008aA00579c1307B0EF2c499aD98a8ce58e58': 'USDT-USD',
            '0x8700dAec35aF8Ff88c16BdF0418774CB3D7599B4': 'SNX-USD',
            '0x4200000000000000000000000000000000000042': 'OP-USD', 
            'ETH': 'ETH-USD',
            'WETH': 'ETH-USD',
            'WBTC': 'BTC-USD',
            'USDC': 'USDC-USD',
            'DAI': 'DAI-USD',
            'USDT': 'USDT-USD',
            'SNX': 'SNX-USD',
            'OP': 'OP-USD'
        }
        
    ######################################################################
    # sure there is plenty of helpful things we can do here in the future ;)
    ######################################################################

    # create some type of token list dictionary that can be used to dynamically pull in token addresses and information

class OptimismGoerli:
    """this class represents a rolodex of relevant informatino regarding the Optimism Goerli network, such as contract addresses, token addresses, and abis."""

    # TODO: set these to be dynamically pulled in a trustless fashion
    def __init__(self):
        """Initialize the Optimism Mainnet class."""
        self.network = 'Optimism Goerli'
        self.chain_id = 420
        self.currency = 'ETH'
        self.rpc_url = 'https://goerli.optimism.io'
        self.explorer_url = 'https://goerli-explorer.optimism.io/'

        # set the rubicon contract addresses
        self.market = '0x6cD8666aBB003073e45D69E5b3aa0b0Fe9CDBF91'
        self.router = '0x6aaEd1985a0e011ca82BB5Df8ebd92063134fd7c'
        self.house = '0x1229036F63679B61910CB1463e5BB57f68D19bb2'
        self.pair = '0x9dBf17d518f722B5Aae5573D808B94024b635529'
        self.utility = '0xd282dB449cC64D136b9D9a4399E7e3F133472EaE'
        self.factory = '0x6838dd21aa01Bde8E600d499A95f9AE02f2bB376'

        # set the subgraph query endpoints
        self.rubicon_market_light = 'https://api.thegraph.com/subgraphs/name/denverbaumgartner/rubiconmarket-light-op-goerli'
        # TODO: this is pointing towards the mainnet subgraph, need to decide whether to keep the detailed subgraph or not, its a lot of data to store... thanks for the compute graph
        self.boiler_plate = 'https://api.thegraph.com/subgraphs/name/denverbaumgartner/boilerplateoptimism'
        self.market_aid = 'https://api.thegraph.com/subgraphs/name/denverbaumgartner/optimismgoerlimarketaid'

        # set up some common asset addresses that may be used 
        # TODO: it would be nice to be able to dynamically pull in a token list in a trustless fashion and populate this list
        self.weth = '0x54e63385c13ECbE3B859991eEdad539d9fDa1167'
        self.op = '0xCeE7148028Ff1B08163343794E85883174a61393'
        self.usdc = '0xe432f229521eE954f80C83257485405E3d848d17'
        self.usdt = '0xD70734Ba8101Ec28b38AB15e30Dc9b60E3c6f433'
        self.forrest_coin = '0x45FA7d7b6C954d17141586e1BD63d2e35d3e26De'

        self.coinbase_tickers = {
            '0x54e63385c13ECbE3B859991eEdad539d9fDa1167': 'ETH-USD',
            '0xe432f229521eE954f80C83257485405E3d848d17': 'USDC-USD',
            '0xD70734Ba8101Ec28b38AB15e30Dc9b60E3c6f433': 'USDT-USD',
            '0xCeE7148028Ff1B08163343794E85883174a61393': 'OP-USD', 
            '0x45FA7d7b6C954d17141586e1BD63d2e35d3e26De': 'MKR-USD',

            # add in the lower case versions of the tokens
            '0x54e63385c13ecbe3b859991eedad539d9fda1167': 'ETH-USD',
            '0xe432f229521ee954f80c83257485405e3d848d17': 'USDC-USD',
            '0xd70734ba8101ec28b38ab15e30dc9b60e3c6f433': 'USDT-USD',
            '0xcee7148028ff1b08163343794e85883174a61393': 'OP-USD', 
            '0x45fa7d7b6c954d17141586e1bd63d2e35d3e26de': 'MKR-USD',
            
            # add in the symbols to ticker mappings
            'ETH': 'ETH-USD',
            'WETH': 'ETH-USD',
            'WBTC': 'BTC-USD',
            'USDC': 'USDC-USD',
            'DAI': 'DAI-USD',
            'USDT': 'USDT-USD',
            'SNX': 'SNX-USD',
            'OP': 'OP-USD',
            'F': 'MKR-USD'
        }
        
        ######################################################################
        # sure there is plenty of helpful things we can do here in the future ;)
        ######################################################################

# set a dictionary that maps the chain id to the network class
networks = {
    10: OptimismMainnet,
    420: OptimismGoerli
}


import json
import os
from enum import Enum
from typing import Optional
from pkg_resources import resource_exists, resource_stream

import yaml
from eth_typing import ChecksumAddress
from web3 import Web3


class NetworkId(Enum):
    # MAINNET
    OPTIMISM = 10

    # TESTNET
    OPTIMISM_GOERLI = 420
    ARBITRUM_GOERLI = 421613
    POLYGON_MUMBAI = 80001


class Network:
    """This class represents a network and can be used to instantiate contracts using the `from_network` method. It
    provides and easy way to configure a client or contracts.
    """

    def __init__(
        self,
        path: str,
        w3: Web3,
        # The below all come from network_config/{network_name}/network.yaml
        name: str,
        chain_id: int,
        currency: str,
        rpc_url: str,
        explorer_url: str,
        rubicon: dict,
        token_addresses: dict,
        # optional custom token config file from the user
        custom_token_addresses_file: Optional[str] = None
    ):
        """Initializes a Network instance.

        :param path: The path to the network configuration.
        :type path: str
        :param w3: The Web3 instance connected to the network.
        :type w3: Web3
        :param name: The name of the network.
        :type name: str
        :param chain_id: The chain ID of the network.
        :type chain_id: int
        :param currency: The base currency of the network.
        :type currency: str
        :param rpc_url: The RPC URL of the network.
        :type rpc_url: str
        :param explorer_url: The URL of the network explorer.
        :type explorer_url: str
        :param rubicon: Dictionary containing Rubicon contract parameters.
        :type rubicon: dict
        :param token_addresses: Dictionary containing token addresses on the network.
        :type token_addresses: dict
        :param custom_token_addresses_file: The name of a yaml file (relative to the current working directory) with
            custom token addresses. Overwrites the token config found in network_config/{chain}/network.yaml.
            (optional, default is None).
        :type custom_token_addresses_file: Optional[str]
        """
        self.name = name
        self.chain_id = chain_id
        self.w3 = w3

        self.currency = currency
        self.rpc_url = rpc_url
        self.explorer_url = explorer_url
        self.rubicon = RubiconContracts(path=path, w3=self.w3, **rubicon)

        checksummed_token_addresses: dict[str, ChecksumAddress] = {}
        for k, v in token_addresses.items():
            checksummed_token_addresses[k] = self.w3.to_checksum_address(v)

        if custom_token_addresses_file:
            working_directory = os.getcwd()

            if not (custom_token_addresses_file.endswith(".yaml") or custom_token_addresses_file.endswith(".yml")):
                raise Exception(f"token_config_file: {custom_token_addresses_file} must be a yaml file.")

            try:
                with open(f"{working_directory}/{custom_token_addresses_file}") as f:
                    custom_token_addresses = yaml.safe_load(f)
                    for k, v in custom_token_addresses.items():
                        checksummed_token_addresses[k] = self.w3.to_checksum_address(v)
            except FileNotFoundError:
                raise Exception(f"could not find token_config_file, expected it to be a yaml file here: "
                                f"{working_directory}/{custom_token_addresses_file}.")
            except AttributeError:
                raise Exception(f"{custom_token_addresses_file} cannot be empty, should be in the format: "
                                f"token_symbol: address, e.g. WETH: chain_specific_weth_address")

        self.token_addresses = checksummed_token_addresses

    @classmethod
    def from_config(cls, http_node_url: str, custom_token_addresses_file: Optional[str] = None) -> "Network":
        """Create a Network instance based on the node url provided. A call is then made to this node to get the
        chain_id which links to network_config/{network_name}/ using the NetworkId Enum.

        :param http_node_url: The URL of the HTTP node for the network.
        :type http_node_url: str
        :param custom_token_addresses_file: The name of a yaml file (relative to the current working directory) with
            custom token addresses. Overwrites the token config found in network_config/{chain}/network.yaml.
            (optional, default is None).
        :type custom_token_addresses_file: Optional[str]
        :return: A Network instance based on the network configuration.
        :rtype: Network
        :raises Exception: If no network configuration file is found for the specified network name.
        """
        w3 = Web3(Web3.HTTPProvider(http_node_url))

        network_name = NetworkId(w3.eth.chain_id).name.lower()

        resource_path = f"../network_config/{network_name}/network.yaml"
        if not resource_exists('rubi', resource_path):
            raise Exception(f"no network config found for {network_name}, there should be a corresponding folder in "
                            f"the network_config directory")
        with resource_stream('rubi', resource_path) as f:
            network_data = yaml.safe_load(f)
            print(f"loaded network config for {network_name} from {resource_path}")

        return cls(path=resource_path, w3=w3, custom_token_addresses_file=custom_token_addresses_file, **network_data)

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class RubiconContracts:
    """This class is a simple wrapper for all the Rubicon contracts on a network. Right now it only contains the market
    and router contracts.
    """

    def __init__(self, path: str, w3: Web3, market: dict, router: dict) -> None:
        """Initialize a RubiconContracts instance.

        :param path: The path to the network configuration.
        :type path: str
        :param w3: The Web3 instance connected to the network.
        :type w3: Web3
        :param market: Dictionary containing market contract information.
        :type market: dict
        :param router: Dictionary containing router contract information.
        :type router: dict
        """
        self.market = ContractRepr(path=path, w3=w3, name="market", **market)
        self.router = ContractRepr(path=path, w3=w3, name="router", **router)

    def __repr__(self):
        items = ("{}={!r}".format(k, self.__dict__[k]) for k in self.__dict__)
        return "{}({})".format(type(self).__name__, ", ".join(items))


class ContractRepr:
    """This class represents a contract on a specific network.
    """

    def __init__(self, path: str, w3: Web3, name: str, address: str) -> None:
        """Initialize a ContractRepr instance which is a representation of a contract containing the contract address
        and abi.

        :param path: The path to the network configuration.
        :type path: str
        :param w3: The Web3 instance connected to the network.
        :type w3: Web3
        :param name: The name of the contract.
        :type name: str
        :param address: The address of the contract.
        :type address: str
        """
        self.address = w3.to_checksum_address(address)

        base_dir_path = os.path.dirname(path)
        file_path = os.path.join(base_dir_path, "abis", f"{name}.json")

        try:
            with resource_stream('rubi', file_path) as f:
                self.abi = json.load(f)
        except FileNotFoundError:
            self.abi = None

    def __repr__(self):
        return f"{type(self).__name__}(address='{self.address}', abi='{'Loaded' if self.abi is not None else 'None'}')"



import time
import pandas as pd
from .helper import Process
from ..sources import AidData, SuperAidData
from ..sources.helper import Gas, Price, networks

class AidProcessing: 
    """this class is used to process the data from the aid datasource"""

    def __init__(self, subgrounds, chain_id, AidData):
        """constructor method to initialize the data source class"""

        self.price = Price()
        self.process = Process()
        self.network = networks[chain_id]()
        self.market_aid = AidData 

    def recreate_aid_history(self, aid, bin_size = 60, raw=False): 
        """this method takes an aid address and builds a dataframe of that aid balance over time. its precision is to the second. it does this by taking the ending assset balance at each second.
        
        :param aid: the address of the aid instance
        :type aid: str
        :param bin_size: the bin size of the analysis (for now this only adjusts the price data precision)
        :type bin_size: int
        :return: a dataframe that shows the various asset balances of the aid instance over a period of time
        :rtype: pandas.DataFrame (columns: )
        """

        # get the aid history
        df = self.market_aid.get_aid_history(aid = aid, bin_size = bin_size)

        # create a dataframe the encompasses the period of the aid history 
        min_timestamp = int(df['timestamp'].min())
        max_timestamp = int(df['timestamp'].max())
        timestamp_range = list(range(min_timestamp, max_timestamp))
        history = pd.DataFrame(timestamp_range, columns=['timestamp'])

        # group by asset 
        aid_history_grouped = df.groupby('asset')

        # iterate through each group and get the balance at the max index position for every timestamp
        for asset, group in aid_history_grouped:
            
            # make sure the group is sorted by: block, block_index, timestamp
            group = group.sort_values(by = ['block', 'block_index', 'timestamp']).reset_index(drop = True)

            # set this sorted index as a column so we can identify the end balance of an asset 
            group['sorted_index'] = group.index

            # get the sum of the credits and debits for each timestamp
            credits = group.groupby(['timestamp']).agg({'credits_debits' : 'sum'}).reset_index()
            credits_raw = group.groupby(['timestamp']).agg({'credits_debits_raw' : 'sum'}).reset_index()

            # fill in the credits and debits for each timestamp
            history[f'{asset}_credits'] = history['timestamp'].map(credits.set_index('timestamp')['credits_debits'].to_dict()).fillna(0)
            history[f'{asset}_credits_raw'] = history['timestamp'].map(credits_raw.set_index('timestamp')['credits_debits_raw'].to_dict()).fillna(0)

            # go through and get the max sorted_index for each timestamp
            idx = group.groupby(['timestamp'])['sorted_index'].idxmax()
            filtered_group = group.loc[idx]

            # fill in the balance for each asset at the end of each timestamp
            history[f'{asset}_balance'] = history['timestamp'].map(filtered_group.set_index('timestamp')['balance'].to_dict()).fillna(method='ffill').fillna(0)
            history[f'{asset}_balance_raw'] = history['timestamp'].map(filtered_group.set_index('timestamp')['balance_raw'].to_dict()).fillna(method='ffill').fillna(0)

            # get coinbase ticker for the pair and retrieve the price data
            ticker = self.network.coinbase_tickers[asset]
            prices = self.price.get_price_in_range(start = min_timestamp, end = max_timestamp, pair = ticker)

            # fill in the price for each asset at the beginning of each minute period
            history[f'{asset}_price'] = history['timestamp'].map(prices).fillna(method='ffill').fillna(method='bfill').fillna(0)

        assets = [asset for asset, group in aid_history_grouped]

        return {'data' : history, 'assets' : assets}

    def evaluate_aid_performance(self, data, assets, start_time=None, end_time=None):
        """this function simply takes an aid history, truncates the dataframe by a start and end time (if given), and compares the book value of holding the asset balance at the beginning of the period (adjusting for any deposits / withdrawals during the period) against the actual aid balances. this function is intended to be used with dataframes created by the function `recreate_aid_history`
        
        :param data: a dataframe (ideally returned by the function recreate_aid_history) 
        :type data: pandas.DataFrame
        :param assets: an array of relevant assests for the dataframe (ideally returned by the function recreate_aid_history) 
        :type assets: array 
        :param start_time: the start time of which to conduct an analysis for (this determines what asset mix is used in comparison against the aids real book value / real asset mix)
        :type start_time: int
        :param end_time: the end time of which to conduct the analysis through 
        :param end_time: int
        """

        if start_time:
            data = data[data['timestamp'] >= start_time].reset_index(drop = True)
        if end_time: 
            data = data[data['timestamp'] <= end_time].reset_index(drop = True)

        # set the credits equal to the balance of each asset at index 0
        for asset in assets:
            data[f'{asset}_credits'].iloc[0] = data[f'{asset}_balance'].iloc[0]
            data[f'{asset}_credits_raw'].iloc[0] = data[f'{asset}_balance_raw'].iloc[0]

            # create a cumulative sum of the credits and debits
            data[f'{asset}_credits_total'] = data[f'{asset}_credits'].cumsum()
            data[f'{asset}_credits_total_raw'] = data[f'{asset}_credits_raw'].cumsum()

            # calculate the delta from the credits totals
            data[f'{asset}_balance_delta'] = data[f'{asset}_balance'] - data[f'{asset}_credits_total']
            data[f'{asset}_balance_delta_raw'] = data[f'{asset}_balance_raw'] - data[f'{asset}_credits_total_raw']

            # calculate the usd value of the balance, credits, and delta
            data[f'{asset}_balance_usd'] = data[f'{asset}_balance'] * data[f'{asset}_price']
            data[f'{asset}_credits_total_usd'] = data[f'{asset}_credits_total'] * data[f'{asset}_price']
            data[f'{asset}_balance_delta_usd'] = data[f'{asset}_balance_delta'] * data[f'{asset}_price']

        data['total_balance_usd'] = data[[f'{asset}_balance_usd' for asset in assets]].sum(axis = 1)
        data['total_credits_usd'] = data[[f'{asset}_credits_total_usd' for asset in assets]].sum(axis = 1)
        data['total_balance_delta_usd'] = data[[f'{asset}_balance_delta_usd' for asset in assets]].sum(axis = 1) 

        return {'data' : data, 'assets' : assets}

    def analyze_aid_history(self, aid, start_time=None, end_time=None, bin_size=60, raw=False):
        """this method takes the aid history entities and composes a dataframe that shows the various asset balances of the aid instance over a period of time, the entire aid history if no start and end time are provided. one caveat, when a dataframe begins after the aid instance was created, the first row has a credit of the balance at that point in time.
        
        :param aid: the address of the aid instance
        :type aid: str
        :param start_time: the start time of the analysis
        :type start_time: int
        :param end_time: the end time of the analysis
        :type end_time: int
        :param bin_size: the bin size of the analysis
        :type bin_size: int
        :return: a dataframe that shows the various asset balances of the aid instance over a period of time
        :rtype: pandas.DataFrame (columns: )
        """

        # get the aid history
        df = self.market_aid.get_aid_history(aid = aid, start_time = start_time, end_time = end_time, bin_size = bin_size)

        # get the unique assets and the tickers needed to retrieve coinbase price data
        assets = list(df['asset'].unique())
        tickers = [self.network.coinbase_tickers[asset] for asset in assets]

        # get the time range
        min_timestamp = int(df['timestamp'].min())
        max_timestamp = int(df['timestamp'].max())
        timestamp_range = list(range(min_timestamp, max_timestamp))

        # get the price data
        price_data = {}
        for asset, ticker in zip(assets, tickers):
            price_data[asset] = self.price.get_price_in_range(start = min_timestamp, end = max_timestamp, pair = ticker)

        # go through each asset and get the relevant balance and price data
        df_grouped = df.groupby('asset', as_index=False)
        asset_deltas = {}
        asset_balances = {}
        credits_debits = {}
        for asset, group in df_grouped:
            
            if group.empty:
                print(f'no data for {asset}')
                continue

            # reset the index
            group = group.reset_index(drop=True)

            # set the initial value as a credit to the account for any previous activity 
            group.loc[0, 'credits_debits'] = group.loc[0, 'balance']
            group.loc[0, 'credis_debits_raw'] = group.loc[0, 'balance_raw']

            # now, within the same column, create the cumulative sum of the credits and debits
            group['credits_debits'] = group['credits_debits'].cumsum()
            group['credits_debits_raw'] = group['credits_debits_raw'].cumsum()

            # now, we can calculate the delta of the credits_debits and credits_debits_raw columns
            group['delta'] = group['balance'] - group['credits_debits'] 
            group['delta_raw'] = group['balance_raw'] - group['credits_debits_raw']

            # now we get the data at each timestamp
            if raw: 
                asset_deltas[asset] = group.set_index('timestamp').to_dict()['delta_raw']
                asset_balances[asset] = group.set_index('timestamp').to_dict()['balance_raw']
                credits_debits[asset] = group.set_index('timestamp').to_dict()['credits_debits_raw']
            else:
                asset_deltas[asset] = group.set_index('timestamp').to_dict()['delta']
                asset_balances[asset] = group.set_index('timestamp').to_dict()['balance']
                credits_debits[asset] = group.set_index('timestamp').to_dict()['credits_debits']

        # build the dataframe
        history = pd.DataFrame(timestamp_range, columns=['timestamp'])
    
        # add in the data for each asset
        for asset in assets:
            history[f'{asset}_balance'] = history['timestamp'].map(asset_balances[asset]).fillna(method='ffill').fillna(0)

        for asset in assets: 
            history[f'{asset}_hodl_balance'] = history['timestamp'].map(credits_debits[asset]).fillna(method='ffill').fillna(0)
        
        for asset in assets:
            history[f'{asset}_delta'] = history['timestamp'].map(asset_deltas[asset]).fillna(method='ffill').fillna(0)

        # add in the price data
        for asset in assets:
            history[f'{asset}_price'] = history['timestamp'].map(price_data[asset]).fillna(method='ffill').fillna(method='bfill')

        # compute values based on the price data
        for asset in assets:
            history[f'{asset}_balance_usd'] = history[f'{asset}_balance'] * history[f'{asset}_price']
        
        for asset in assets:
            history[f'{asset}_hodl_balance_usd'] = history[f'{asset}_hodl_balance'] * history[f'{asset}_price']

        for asset in assets:
            history[f'{asset}_delta_usd'] = history[f'{asset}_delta'] * history[f'{asset}_price']

        # compute the total balance and delta
        history['total_balance_usd'] = history[[f'{asset}_balance_usd' for asset in assets]].sum(axis=1)
        history['total_hodl_balance_usd'] = history[[f'{asset}_hodl_balance_usd' for asset in assets]].sum(axis=1)
        history['total_delta_usd'] = history[[f'{asset}_delta_usd' for asset in assets]].sum(axis=1)
        
        return {'data' : history, 'tokens' : assets, 'tickers' : tickers}
    
    def idealized_mix_analysis(self, data, tokens, asset_mix, rebalance_interval = 60):
        """this function is intended so simulate the performance of a market aid instance against its "idealized mix", that is, a portfolio holding the same assets in a pre-defined proportion. the rebalancing of the idealized mix is done on an interval basis 

        :param data: the dataframe containing the data to analyze. pulled from the analyze_aid_history function
        :type data: pd.DataFrame
        :param tokens: the assets that the aid held during the period
        :type tokens: list
        :param asset_mix: the idealized mix of the assets held by the aid, a dictionary with the asset as the key and the proportion as the value (i.e. {'eth' : 0.5, 'usdc' : 0.5})
        :type asset_mix: dict
        """

        # sanity check and ensure the data is timestamp sorted in ascending order
        data = data.sort_values('timestamp', ascending=True)
        data.reset_index(inplace=True)

        # iterate through the data and compute the idealized mix at every interval
        for index, row in data.iterrows():

            if index == 0:
                usd_balance = row['total_balance_usd']

                # set the initial idealized mix 
                for token in tokens: 
                    return 'TODO'
        #for token in tokens: 
        #    data[f'{token}_idealized_mix_balance_usd'] = data[f'{token}_balance_usd'] * asset_mix[token]

        #return TODO

    def aid_performance_evaluation(self, data, tokens, asset_mix = None):
        """this function is intended to take a dataframe with a set of tokens and determine the performance of the aid over the given time period. it does this in two ways: 
            1. hodl strat comp: compare the performance of aid activity over the period to the performance of simply holding the assets (in the same proportions) as the beginning of the period
            2. asset mix comp: compare the performance of the aid activity to the performance of the holding the assets at a set proprtion (asset mix) as determined by the beginning of the period
        """

        # sanity check and ensure the data is timestamp sorted in ascending order
        data = data.sort_values('timestamp', ascending=True)
        data.reset_index(inplace=True)

        # get the initial asset balances at the beginning of the period
        initial_asset_balances = {}
        data['total_balance_initial_usd'] = 0
        for token in tokens:
            initial_asset_balances[token] = data[f'{token}_balance'].iloc[0]
            data[f'{token}_balance_initial'] = initial_asset_balances[token]
            data[f'{token}_balance_initial_usd'] = data[f'{token}_balance_initial'] * data[f'{token}_price']
            data['total_balance_initial_usd'] += data[f'{token}_balance_initial_usd']
        
        # compute the hodl strat performance delta 
        data['hodl_strat_performance_delta'] = data['total_balance_usd'] - data['total_balance_initial_usd']

        # if there is an asset mix, get the initial prices as well and compute the specific asset mix
        if asset_mix: 
            initial_usd_balance = data['total_balance_usd'].iloc[0]
            initial_asset_prices = {}
            ideal_asset_balances = {}
            data['total_balance_ideal_usd'] = 0
            for token in tokens:
                initial_asset_prices[token] = data[f'{token}_price'].iloc[0]
                ideal_asset_balances[token] = (initial_usd_balance) * asset_mix[token] / initial_asset_prices[token]
                data[f'{token}_balance_ideal'] = ideal_asset_balances[token]
                data[f'{token}_balance_ideal_usd'] = data[f'{token}_balance_ideal'] * data[f'{token}_price']
                data['total_balance_ideal_usd'] += data[f'{token}_balance_ideal_usd']
            
            # compute the asset mix strat performance delta
            data['asset_mix_strat_performance_delta'] = data['total_balance_usd'] - data['total_balance_ideal_usd']

        return data

    def aid_fill_tracking(self, aid, asset, quote, start_time=None, end_time=None, bin_size=60, first=1000000000):
        """this function is intended to track the fills of an aid contract over time. specifically, it takes the offers for a set pair and tracks the fill the occurred within
        a series of "time bins" (this can be as granular as a second and arbitrarily large (seconds)) for a set pair. this is denominated by an asset and quote, with the 
        denomination deciding the direction of trades
         
        :param aid: the aid contract to track
        :type aid: str
        :param asset: the asset to track
        :type asset: str
        :param quote: the quote to track 
        :type quote: str
        :param start_time: the start time of the tracking period
        :type start_time: int
        :param end_time: the end time of the tracking period
        :type end_time: int
        :param bin_size: the size of the time bins
        :type bin_size: int
        :param first: the amount of offers to get from the contract
        """

        # get the offers from the contract
        # TODO: this needs to be updated to add additional filtering as soon as its supported in the low level api
        offers = self.market_aid.get_aid_offers(aid.lower())
        
        # if the dataframe is empty, return an empty dataframe with the correct columns
        if len(offers) == 0:
            return pd.DataFrame(columns=['time_bin', 'direction', 'offers_pay_amt_formatted', 'offers_paid_amt_formatted', 'offers_buy_amt_formatted', 'offers_bought_amt_formatted', 'price', 'offers_live'])

        # filter the offers for the desired asset quote pair, the offers_pay_gem_id and offers_buy_gem_id can be either the asset or the quote, so we need to check both
        if len(quote) < 6:
            asks = offers[(offers['offers_pay_gem_symbol'] == asset.upper()) & (offers['offers_buy_gem_symbol'] == quote.upper())]
            bids = offers[(offers['offers_pay_gem_symbol'] == quote.upper()) & (offers['offers_buy_gem_symbol'] == asset.upper())]
        else:
            asks = offers[(offers['offers_pay_gem_id'] == asset.lower()) & (offers['offers_buy_gem_id'] == quote.lower())]
            bids = offers[(offers['offers_pay_gem_id'] == quote.lower()) & (offers['offers_buy_gem_id'] == asset.lower())]

        # group the offers by the time bin they occurred in
        asks['time_bin'] = asks['offers_transaction_timestamp'].apply(lambda x: int((x // bin_size) * bin_size))
        bids['time_bin'] = bids['offers_transaction_timestamp'].apply(lambda x: int((x // bin_size) * bin_size))

        # group the offers by the time bin they occurred in
        asks = asks.groupby('time_bin').sum(numeric_only=True)
        bids = bids.groupby('time_bin').sum(numeric_only=True)

        # add a direction column 
        asks['direction'] = 'ask' # selling the asset
        bids['direction'] = 'bid' # buying the asset

        # transform the index into a column
        asks = asks.reset_index().rename(columns={'index': 'time_bin'})
        bids = bids.reset_index().rename(columns={'index': 'time_bin'})

        # calculate the ask and bid prices 
        asks['price'] =  asks['offers_buy_amt_formatted'] / asks['offers_pay_amt_formatted'] # price to buy the asset from the seller
        bids['price'] =  bids['offers_pay_amt_formatted'] / bids['offers_buy_amt_formatted'] # price to sell the asset to the buyer
        
        # subset the data to the relevant information
        asks = asks[['time_bin', 'direction', 'offers_pay_amt_formatted', 'offers_paid_amt_formatted', 'offers_buy_amt_formatted', 'offers_bought_amt_formatted', 'price', 'offers_live']]
        bids = bids[['time_bin', 'direction', 'offers_pay_amt_formatted', 'offers_paid_amt_formatted', 'offers_buy_amt_formatted', 'offers_bought_amt_formatted', 'price', 'offers_live']]
        
        df = pd.concat([asks, bids]).sort_values('time_bin', ascending=True)
        df.reset_index(inplace=True)
        
        return df

        


import logging as log
from threading import Thread
from time import sleep
from typing import Optional, Callable, Type, Dict, Any

from eth_account.datastructures import SignedTransaction
from eth_typing import ChecksumAddress
from web3 import Web3
from web3._utils.filters import LogFilter  # noqa
from web3.contract import Contract
from web3.contract.contract import ContractFunction
from web3.types import ABI, Nonce

from rubi.contracts.types.events import BaseEvent


class BaseContract:
    """Base class representation of a contract which defines the structure of a contract and provides several helpful
    methods that can be used by subclass contracts that extend this contract.

    :param w3: Web3 instance
    :type w3: Web3
    :param contract: Contract instance
    :type contract: Contract
    :param wallet: a wallet address of the signer (optional, default is None)
    :type wallet: Optional[ChecksumAddress]
    :param key: the private key of the signer (optional, default is None)
    :type key: Optional[str]
    """

    def __init__(
        self,
        w3: Web3,
        contract: Contract,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ):
        """constructor method"""
        if (wallet is None) != (key is None):
            raise Exception("both a wallet and a key are required to sign transactions. provide both or omit both")

        self.contract = contract
        self.address = contract.address
        self.w3 = w3
        self.chain_id = self.w3.eth.chain_id

        # Signing permissions
        self.signing_permissions = (wallet is not None and key is not None)

        if self.signing_permissions:
            log.info(f"instantiated {self.__class__} with signing rights")

            # Force typing as my editors inspection is throwing a tantrum
            self.wallet = wallet  # type: ChecksumAddress
            self.key = key  # type: str

    @classmethod
    def from_address_and_abi(
        cls,
        w3: Web3,
        address: ChecksumAddress,
        contract_abi: ABI,
        wallet: Optional[ChecksumAddress] = None,
        key: Optional[str] = None
    ) -> "BaseContract":
        """Create a BaseContract instance from the contract address and ABI.

        :param w3: The Web3 instance.
        :type w3: Web3
        :param address: The address of the contract.
        :type address: ChecksumAddress
        :param contract_abi: The ABI of the contract.
        :type contract_abi: ABI
        :param wallet: The wallet address to use for interacting with the contract (optional, default is None).
        :type wallet: Optional[ChecksumAddress]
        :param key: The private key of the wallet (optional, default is None).
        :type key: Optional[str]
        :return: An instance of BaseContract.
        :rtype: BaseContract
        """

        contract = w3.eth.contract(address=address, abi=contract_abi)

        return cls(w3=w3, contract=contract, wallet=wallet, key=key)

    ######################################################################
    # event listeners
    ######################################################################

    # TODO: revisit poll time. Right now it is set to block production time of optimism according to:
    #  https://community.optimism.io/docs/protocol/2-rollup-protocol/#block-storage
    #  however arbitrum produces blocks faster (every 0.25 secs) according to:
    #  https://arbiscan.io/chart/blocktime so we may be prudent to account for different chains
    #  however the better way to do this is probably a websocket connection to the node
    # TODO: investigate using a more generic filter so that you don't need to poll for each event as this could
    #  spam the node that is being connected to
    def start_event_poller(
        self,
        pair_name: str,
        event_type: Type[BaseEvent],
        argument_filters: Optional[Dict[str, Any]] = None,
        event_handler: Optional[Callable] = None,
        poll_time: int = 2
    ) -> None:
        """Start a thread which runs an event poller for a specific event type.

        :param pair_name: The name of the pair we are monitoring events of.
        :type pair_name: str
        :param event_type: The type of event to poll for.
        :type event_type: Type[BaseEvent]
        :param argument_filters: Optional filters that the node will filter events on (optional, default is None).
        :type argument_filters: Optional[Dict[str, Any]]
        :param event_handler: Optional event handler function. Defaults to using the events default handler.
        :type event_handler: Optional[Callable]
        :param poll_time: The time interval between each poll in seconds. Defaults to 2 seconds.
        :type poll_time: int
        """

        event_filter = event_type.create_event_filter(contract=self.contract, argument_filters=argument_filters)
        handler = event_handler if event_handler is not None else event_type.default_handler

        thread = Thread(
            target=self._start_default_event_poller,
            args=(pair_name, event_type, event_filter, handler, poll_time),
            daemon=True
        )
        thread.start()

    @staticmethod
    def _start_default_event_poller(
        pair_name: str,
        event_type: Type[BaseEvent],
        event_filter: LogFilter,
        event_handler: Callable,
        poll_time: int
    ) -> None:
        """Start the default event poller loop. This thread will stop if the pair is removed from the client.

        :param pair_name: The name of the event pair.
        :type pair_name: str
        :param event_type: The type of the event.
        :type event_type: Type[BaseEvent]
        :param event_filter: The event filter to retrieve new entries.
        :type event_filter: LogFilter
        :param event_handler: The event handler function.
        :type event_handler: Callable
        :param poll_time: The time interval between poll iterations in seconds.
        :type poll_time: int
        """
        polling = True

        while polling:
            try:
                for event_data in event_filter.get_new_entries():
                    event_handler(pair_name, event_type, event_data)
            except Exception as e:
                log.error(e)

                # TODO: this is a hack to detect if a PairDoesNotExistException is raised and polling should stop.
                #  Currently an additional except PairDoesNotExistException as e: cannot be added as this causes a
                #  circular import. Think about restructuring the directories to avoid this (e.g one root level types
                #  directory).
                if "add pair to the client" in str(e):
                    polling = False

            sleep(poll_time)

    ######################################################################
    # helper methods
    ######################################################################

    def _default_transaction_handler(
        self,
        instantiated_contract_function: ContractFunction,
        gas: int,
        nonce: Optional[int] = None,
        max_fee_per_gas: Optional[int] = None,
        max_priority_fee_per_gas: Optional[int] = None
    ) -> str:
        """Default transaction handler for executing transactions against this contract. This function will build, sign
        and execute a transaction with reasonable defaults (mostly from the web3py library).

        Note: if a nonce is not passed then this function will query to the wallet to get the nonce and also wait for
        the transaction receipt before returning.

        :param instantiated_contract_function: The instantiated contract function to call.
        :type instantiated_contract_function: ContractFunction
        :param gas: The gas limit for the transaction.
        :type gas: int
        :param nonce: Optional nonce value for the transaction (optional, default is None).
        :type nonce: Optional[int]
        :param max_fee_per_gas: Optional maximum fee per gas for the transaction (optional, default is None).
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: Optional maximum priority fee per gas for the transaction.
            (optional, default is None).
        :type max_priority_fee_per_gas: Optional[int]
        :return: The transaction hash of the executed transaction.
        :rtype: str
        """
        if not self.signing_permissions:
            raise Exception(f"cannot write transaction without signing rights. "
                            f"re-instantiate {self.__class__} with a wallet and private key")

        base_txn = self._transaction_params(
            gas=gas,
            nonce=nonce,
            max_fee_per_gas=max_fee_per_gas,
            max_priority_fee_per_gas=max_priority_fee_per_gas
        )

        txn = instantiated_contract_function.build_transaction(
            transaction=base_txn
        )

        signed_txn = self.w3.eth.account.sign_transaction(
            transaction_dict=txn,
            private_key=self.key
        )
        result = self.w3.eth.send_raw_transaction(signed_txn.rawTransaction)

        # If a user is not providing a nonce, wait for the transaction to either be confirmed or rejected before
        # continuing
        if nonce is None:
            self._wait_for_transaction_receipt(transaction=signed_txn)

        return result.hex()

    def _transaction_params(
        self,
        nonce: Optional[Nonce],
        gas: int,
        max_fee_per_gas: Optional[int],
        max_priority_fee_per_gas: Optional[int]
    ) -> Dict:
        """Build transaction parameters Dict for a transaction. If a key is associated with a None value after building
        the Dict then this key will be removed before returning the dict.

        :param nonce: Optional nonce value for the transaction (optional, default is None).
        :type nonce: Optional[Nonce]
        :param gas: The gas limit for the transaction.
        :type gas: int
        :param max_fee_per_gas: Optional maximum fee per gas for the transaction (optional, default is None).
        :type max_fee_per_gas: Optional[int]
        :param max_priority_fee_per_gas: Optional maximum priority fee per gas for the transaction.
            (optional, default is None).
        :type max_priority_fee_per_gas: Optional[int]
        :return: The transaction parameters dictionary.
        :rtype: Dict
        """

        if nonce is None:
            nonce = self.w3.eth.get_transaction_count(self.wallet)

        transaction = {
            'chainId': self.chain_id,
            'gas': gas,
            'maxFeePerGas': max_fee_per_gas,
            'maxPriorityFeePerGas': max_priority_fee_per_gas,
            'nonce': nonce,
        }

        return {key: value for key, value in transaction.items() if value is not None}

    # TODO: actually make use of the transaction receipt to update data structures. This can be useful for example
    #  in getting the amount of gas and cost of gas of the transaction
    def _wait_for_transaction_receipt(self, transaction: SignedTransaction) -> None:
        """Wait for the transaction receipt and check if the transaction was successful.

        :param transaction: The signed transaction object.
        :type transaction: SignedTransaction
        :raises Exception: If the transaction fails.
        """
        if self.w3.eth.wait_for_transaction_receipt(transaction.hash)['status'] == 0:
            raise Exception(f"transaction {transaction.hash.hex()} failed")



import pandas as pd
from ..sources import MarketData
from ..sources.helper import Gas, Price, networks

class User:
    """this class acts as the entry point for accessing data specific to a user or series of users. it is used to create a user data object that can be used to access a variety of data sources and tooling. more to come soon!
    """

    def __init__(self, subgrounds, MarketDataOP):
        """constructor method. creates a subgrounds object that is then used to initialize a variety of data sources and tooling
        
        :param subgrounds: a subgrounds object
        :type subgrounds: subgrounds.Subgrounds
        :param MarketDataOP: a market data object for the optimism network
        :type MarketDataOP: rubi.data.sources.market.MarketData
        """
        
        self.network = networks
        self.subgrounds = subgrounds
        self.market_optimism = MarketDataOP

    def get_user_trading_transactions(self, user, pair = None, filled = None, cancelled = None, live = None, pay_gem = None, buy_gem = None, start_time = None, end_time = None, first = 1000000000):
        # TODO: all aggregate functions will need to be modified in a multi-chain world :soon: hehe ;)
        """this function is used to return a dataframe of a user's trading activity transactions. it returns a dataframe containing the user's offers and takes transactions on the market.
        
        :param user: the user's wallet address
        :type user: str
        :param pair: the address of the pair of the offer, defaults to None. direction of the pair matters and corresponds to the array passed in the following order: [pay_gem, buy_gem]. defaults to None
        :type pair: array, optional
        :param filled: whether the offer has been filled, defaults to None
        :type filled: bool, optional
        :param cancelled: whether the offer has been cancelled, defaults to None
        :type cancelled: bool, optional
        :param live: whether the offer is live, defaults to None
        :type live: bool, optional
        :param pay_gem: the address of the pay_gem of the offer, defaults to None
        :type pay_gem: str, optional
        :param buy_gem: the address of the buy_gem of the offer, defaults to None
        :type buy_gem: str, optional
        :param start_time: the start time of the offer, defaults to None
        :type start_time: int, optional
        :param end_time: the end time of the offer, defaults to None
        :type end_time: int, optional
        :return: a dataframe containing the user's offers and takes
        :rtype: pandas.DataFrame
        """

        op_offers = self.market_optimism.get_offers(maker = user, pair = pair, filled = filled, cancelled = cancelled, live = live, pay_gem = pay_gem, buy_gem = buy_gem, start_time = start_time, end_time = end_time, first = first)
        op_trades = self.market_optimism.get_trades(taker = user, pair = pair, pay_gem = pay_gem, buy_gem = buy_gem, start_time = start_time, end_time = end_time, first = first)

        if op_offers.empty: 
            op_offers = pd.DataFrame(columns=['offers_maker_id', 'offers_transaction_id', 'offers_transaction_timestamp'])
        if op_trades.empty:
            op_trades = pd.DataFrame(columns=['takes_taker_id', 'takes_transaction_id', 'takes_transaction_timestamp'])

        # transform the dataframe column names so they can be merged
        #df.rename(columns={'A': 'a', 'B': 'b', 'C': 'c'}, inplace=True)
        op_offers.rename(columns={
            'offers_maker_id' : 'user',
            'offers_transaction_id' : 'transaction_hash',
            'offers_transaction_timestamp' : 'transaction_timestamp'
        }, inplace=True)

        op_trades.rename(columns={
            'takes_taker_id' : 'user',
            'takes_transaction_id' : 'transaction_hash',
            'takes_transaction_timestamp' : 'transaction_timestamp'
        }, inplace=True)

        # add a column that labels the transaction as an offer or a trade
        op_offers = op_offers.assign(transaction_type='offer')
        op_trades = op_trades.assign(transaction_type='trade')

        # subset the offers and trades dataframes to only include the relevant columns
        op_offers = op_offers[['user', 'transaction_hash', 'transaction_timestamp', 'transaction_type']]
        op_trades = op_trades[['user', 'transaction_hash', 'transaction_timestamp', 'transaction_type']]

        # merge the offers and trades dataframes
        df = pd.concat([op_offers, op_trades])
        # TODO: reset the index
        df.drop_duplicates(inplace=True)

        return df

class SuperUser(User): 
    """this class acts as an extension of the user class with additional functionality that is enabled by being connected to a node. 
    """

    def __init__(self, w3, subgrounds, MarketDataOP):
        """constructor method. creates a subgrounds object that is then used to initialize a variety of data sources and tooling
        
        :param w3: a web3 object
        :type w3: web3.main.Web3
        :param subgrounds: a subgrounds object
        :type subgrounds: subgrounds.Subgrounds
        :param MarketDataOP: a market data object for the optimism network
        :type MarketDataOP: rubi.data.sources.market.MarketData
        """
        
        # initialize the user class
        super().__init__(subgrounds, MarketDataOP)

        # set class variables
        self.w3 = w3
        self.gas = Gas(self.w3)

    def get_user_trading_gas_spend(self, user, pair = None, filled = None, cancelled = None, live = None, pay_gem = None, buy_gem = None, start_time = None, end_time = None, first = 1000000000,
        total_fee_eth = True, total_fee_usd = True, l2_gas_price = False, l2_gas_used = False, l1_gas_used = False, l1_gas_price = False, l1_fee_scalar = False, l1_fee = False, l2_fee = False, total_fee = False, l1_fee_eth = False, l2_fee_eth = False, eth_price = False, l1_fee_usd = False, l2_fee_usd = False):
        """this function is used to return a dataframe of a user's trading activity transactions. it returns a dataframe containing the user's offers and takes transactions on the market, along with relevant gas data for the transactions.
        
        :param user: the user's wallet address
        :type user: str
        :param pair: the address of the pair of the offer, defaults to None. direction of the pair matters and corresponds to the array passed in the following order: [pay_gem, buy_gem]. defaults to None
        :type pair: array, optional
        :param filled: whether the offer has been filled, defaults to None
        :type filled: bool, optional
        :param cancelled: whether the offer has been cancelled, defaults to None
        :type cancelled: bool, optional
        :param live: whether the offer is live, defaults to None
        :type live: bool, optional
        :param pay_gem: the address of the pay_gem of the offer, defaults to None
        :type pay_gem: str, optional
        :param buy_gem: the address of the buy_gem of the offer, defaults to None
        :type buy_gem: str, optional
        :param start_time: the start time of the offer, defaults to None
        :type start_time: int, optional
        :param end_time: the end time of the offer, defaults to None
        :type end_time: int, optional
        :param first: the number of transactions to return, defaults to 1000000000
        :type first: int, optional
        :param total_fee_eth: whether to include the total fee in eth, defaults to True
        :type total_fee_eth: bool, optional
        :param total_fee_usd: whether to include the total fee in usd, defaults to True
        :type total_fee_usd: bool, optional
        :param l2_gas_price: whether to include the l2 gas price, defaults to False
        :type l2_gas_price: bool, optional
        :param l2_gas_used: whether to include the l2 gas used, defaults to False
        :type l2_gas_used: bool, optional
        :param l1_gas_used: whether to include the l1 gas used, defaults to False
        :type l1_gas_used: bool, optional
        :param l1_gas_price: whether to include the l1 gas price, defaults to False
        :type l1_gas_price: bool, optional
        :param l1_fee_scalar: whether to include the l1 fee scalar, defaults to False
        :type l1_fee_scalar: bool, optional
        :param l1_fee: whether to include the l1 fee, defaults to False
        :type l1_fee: bool, optional
        :param l2_fee: whether to include the l2 fee, defaults to False
        :type l2_fee: bool, optional
        :param total_fee: whether to include the total fee, defaults to False
        :type total_fee: bool, optional
        :param l1_fee_eth: whether to include the l1 fee in eth, defaults to False
        :type l1_fee_eth: bool, optional
        :param l2_fee_eth: whether to include the l2 fee in eth, defaults to False
        :type l2_fee_eth: bool, optional
        :param eth_price: whether to include the eth price, defaults to False
        :type eth_price: bool, optional
        :param l1_fee_usd: whether to include the l1 fee in usd, defaults to False
        :type l1_fee_usd: bool, optional
        :param l2_fee_usd: whether to include the l2 fee in usd, defaults to False
        :type l2_fee_usd: bool, optional
        :return: a dataframe containing the user's offers and takes
        :rtype: pandas.DataFrame
        """

        # get the user's trading transactions
        df = self.get_user_trading_transactions(user, pair = pair, filled = filled, cancelled = cancelled, live = live, pay_gem = pay_gem, buy_gem = buy_gem, start_time = start_time, end_time = end_time, first = first)

        # update the dataframe to contain the relevant gas data
        df = self.gas.txn_dataframe_update(df, 'transaction_hash', 'transaction_timestamp', total_fee_eth = total_fee_eth, total_fee_usd = total_fee_usd, l2_gas_price = l2_gas_price, l2_gas_used = l2_gas_used, l1_gas_used = l1_gas_used, l1_gas_price = l1_gas_price, l1_fee_scalar = l1_fee_scalar, l1_fee = l1_fee, l2_fee = l2_fee, total_fee = total_fee, l1_fee_eth = l1_fee_eth, l2_fee_eth = l2_fee_eth, eth_price = eth_price, l1_fee_usd = l1_fee_usd, l2_fee_usd = l2_fee_usd)

        return df


#from helper import Gas
import pandas as pd
from .helper import Gas
from subgrounds import Subgrounds
from subgrounds.pagination import ShallowStrategy

from rubi.data.sources.helper import networks

class AidData: 
    """this class acts as the main access point to a variety of data and tooling for MarketAid.sol contracts.
    the MarketAid.sol contracts are meant to be a hub for extended functionality for automated traders on the Rubicon protocol.
    the end goal of this class, and this SDK by extension, is to act as a collaborative platform for the advancement of shared data tooling and analysis.
    """

    def __init__(self, subgrounds, chain_id):
        """constructor method

        :param subgrounds: the subgrounds object
        :type subgrounds: subgrounds.Subgrounds
        :param chain_id: the chain id of the network that is of interest
        :type chain_id: int
        """
        self.network = networks[chain_id]()
        self.subgrounds = subgrounds
        self.market_aid = self.subgrounds.load_subgraph(self.network.market_aid)
    
    def get_aid_history(self, aid=None, asset=None, start_time=None, end_time=None, bin_size=60, first=1000000000):
        """this method is used to get all of the asset history for a given aid. it tracks any change in asset balance on the market aid due to a transaction. currently, this is all supported through the event emittals on the market aid contract. any activity not recorded by a covered event will be missed...
        
        :param aid: the address of the market aid that is of interest
        :type aid: str
        :param bin_size: the size of the time bins that are of interest, defaults to 60. this is actually not really used at the moment, so reach out if you have a use case for it
        :type bin_size: int, optional
        :param first: the number of transactions to return, defaults to 1000000000
        :type first: int, optional
        """
        # TODO: we are going to need to extned the information we collect from the subgraph, and possibly modify the subgraph itself, in order to make sure we can account for deposits / withdrawals in any pnl calculation

        AidToken = self.market_aid.AidToken
        AidTokenHistory = self.market_aid.AidTokenHistory

        AidTokenHistory.time_bin = (AidTokenHistory.timestamp // bin_size) * bin_size
        AidTokenHistory.balance_formatted = AidTokenHistory.balance / 10 ** AidTokenHistory.aid_token.token.decimals
        AidTokenHistory.balance_change_formatted = AidTokenHistory.balance_change / 10 ** AidTokenHistory.aid_token.token.decimals

        where = {}
        if aid: 
            where['aid'] = aid.lower()
        if asset:
            where['aid_token_'] = {'token' : asset.lower()}
        if start_time:
            where['timestamp_gte'] = start_time
        if end_time:
            where['timestamp_lte'] = end_time

        #histories = self.market_aid.Query.aidTokenHistories(first = first, where = [AidToken.aid == aid.lower()])
        if where:
            histories = self.market_aid.Query.aidTokenHistories(first = first, where = where)
        else:
            histories = self.market_aid.Query.aidTokenHistories(first = first)

        field_paths = [
            histories.timestamp,
            histories.time_bin,
            histories.aid_token.token.symbol,
            histories.balance_formatted,
            histories.balance_change_formatted,
            histories.transaction.id,
            histories.transaction.block,
            histories.transaction.index,
            histories.index,
            histories.balance,
            histories.balance_change,
            histories.book_update
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)

        # rename the columns if the dataframe is not empty, else return an empty dataframe with the correct columns
        column_name_mapping = {
            'aidTokenHistories_timestamp' : 'timestamp',
            'aidTokenHistories_time_bin' : 'time_bin',
            'aidTokenHistories_aid_token_token_symbol' : 'asset',
            'aidTokenHistories_balance_formatted' : 'balance',
            'aidTokenHistories_balance_change_formatted' : 'balance_change',
            'aidTokenHistories_transaction_id' : 'txn',
            'aidTokenHistories_transaction_block' : 'block',
            'aidTokenHistories_transaction_index' : 'block_index',
            'aidTokenHistories_index' : 'index',
            'aidTokenHistories_balance' : 'balance_raw',
            'aidTokenHistories_balance_change' : 'balance_change_raw',
            'aidTokenHistories_book_update' : 'book_update'
        }

        if df.empty:
            return pd.DataFrame(columns=column_name_mapping.values())
        else:
            df = df.rename(columns=column_name_mapping)
            df = df.sort_values(['block', 'block_index', 'index']).reset_index(drop=True)

            #TODO: this should most likely be moved to a synthetic field going forward 
            df['credits_debits'] = 0
            df['credits_debits_raw'] = 0
            mask = df['book_update'] == True
            df.loc[mask, 'credits_debits'] = df.loc[mask, 'balance_change']
            df.loc[mask, 'credits_debits_raw'] = df.loc[mask, 'balance_change_raw']
            return df

    # TODO: issue #19 dealing with filtering nested entities, that or modify the subgraph so that we can filter by timestamp
    def get_aid_offers(self, aid=None, pair=None, pay_gem=None, buy_gem=None, start_time=None, end_time=None, first=1000000000):
        # TODO: figure out if we want to filter by token symbol or token address, or both
        # address is going to be network specific, but if we are losing the ability to use the symbol through the subgraph then we will need to change several things in the SDK :( 
        """this method is used to get all of the offers for a given aid that have been tracked through the market aid subgraph. https://github.com/RubiconDeFi/rubi-subgraphs

        # relevant issue: need to add filtering for pay_gem, buy_gem, start_time, end_time either through added functionality in subgrounds for nested filtering or through the underlying subgraph 
        # i am currently hesistant to do much with expected changes that will occur to the subgraph schema that will necessitate changes here as well
        # if you are reading this, we are moving to event only based subgraphs x0

        :param aid: the address of the market aid that is of interest
        :type aid: str
        :param pay_gem: the address of the pay gem that is of interest
        :type pay_gem: str
        :param buy_gem: the address of the buy gem that is of interest
        :type buy_gem: str
        :param start_time: the start time of the time range that is of interest, defaults to None
        :type start_time: int, optional
        :param end_time: the end time of the time range that is of interest, defaults to None
        :type end_time: int, optional
        :param first: the number of offers to return, defaults to 1000000000
        :type first: int, optional
        """

        Offer = self.market_aid.Offer
        Offer.pay_amt_formatted = Offer.pay_amt / 10 ** Offer.pay_gem.decimals
        Offer.buy_amt_formatted = Offer.buy_amt / 10 ** Offer.buy_gem.decimals
        Offer.paid_amt_formatted = Offer.paid_amt / 10 ** Offer.pay_gem.decimals
        Offer.bought_amt_formatted = Offer.bought_amt / 10 ** Offer.buy_gem.decimals
    
        where = {}
        if aid:
            where['maker'] = aid.lower()
        if pay_gem:
            where['pay_gem'] = pay_gem.lower()
        if buy_gem:
            where['buy_gem'] = buy_gem.lower()
        if start_time:
            where['transaction'] = {'timestamp_gte': start_time}
        if end_time:
            where['transaction'] = {'timestamp_lte': end_time}

        if where:
            offers = self.market_aid.Query.offers(first = first, where = where)
        else:
            offers = self.market_aid.Query.offers(first = first)
        
        field_paths = [
            offers.id,
            offers.transaction.timestamp,
            offers.removed_timestamp,
            offers.pay_gem.symbol,
            offers.pay_amt_formatted,
            offers.paid_amt_formatted,
            offers.buy_gem.symbol,
            offers.buy_amt_formatted,
            offers.bought_amt_formatted,
            offers.pay_gem.id,
            offers.buy_gem.id,
            offers.live
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)

        return df

    def get_aid_txns(self, aid=None, start_time=None, end_time=None, first=1000000000): 
        """this method is used to get all of the transactions for a given aid that have been tracked through the market aid subgraph. https://github.com/RubiconDeFi/rubi-subgraphs
        
        :param aid: the address of the market aid that is of interest, defaults to None
        :type aid: str, optional
        :param start_time: the start time of the time range that is of interest, defaults to None
        :type start_time: int, optional
        :param end_time: the end time of the time range that is of interest, defaults to None
        :type end_time: int, optional
        :param first: the number of transactions to return, defaults to 1000000000
        :type first: int, optional
        """

        Transaction = self.market_aid.Transaction

        where = []
        if aid:
            where.append(Transaction.aid == aid)
        if start_time:
            where.append(Transaction.timestamp >= start_time)
        if end_time:
            where.append(Transaction.timestamp <= end_time)

        if where == []:
            txns = self.market_aid.Query.transactions(first=first)
        else:
            txns = self.market_aid.Query.transactions(first=first, where=where)

        field_paths = [
            txns.id,
            txns.timestamp,
            txns.aid.id
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)

        return df

    def get_book_updates(self, aid=None, asset=None, start_time=None, end_time=None, first=1000000000):
        """this method is used to get all of the book updates for a given aid that have been tracked through the market aid subgraph.
        
        :param aid: the address of the market aid that is of interest, defaults to None
        :type aid: str, optional
        :param asset: the address of the asset that is of interest, defaults to None
        :type asset: str, optional
        :param start_time: the start time of the time range that is of interest, defaults to None
        :type start_time: int, optional
        :param end_time: the end time of the time range that is of interest, defaults to None
        :type end_time: int, optional
        :param first: the number of book updates to return, defaults to 1000000000
        :type first: int, optional
        :return: a dataframe containing the book updates
        :rtype: pandas.DataFrame
        """

        BookUpdate = self.market_aid.BookUpdate
        BookUpdate.amount_formatted = BookUpdate.amount / 10 ** BookUpdate.aid_token.token.decimals

        where = []
        if aid:
            where.append(BookUpdate.aid == aid.lower())
        #if asset:
        #    where.append(BookUpdate.asset == asset)
        if start_time:
            where.append(BookUpdate.timestamp >= start_time)
        if end_time:
            where.append(BookUpdate.timestamp <= end_time)

        if where == []:
            book_updates = self.market_aid.Query.bookUpdates(first=first)
        else:
            book_updates = self.market_aid.Query.bookUpdates(first=first, where=where)

        field_paths = [
            book_updates.timestamp,
            book_updates.aid.id,
            book_updates.aid_token.token.symbol,
            book_updates.amount,
            book_updates.amount_formatted,
            book_updates.transaction.id,
            book_updates.transaction.block,
            book_updates.transaction.index,
            book_updates.index
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)

        # rename the columns: timestamp, aid, asset, txn, amount, amount_formatted
        column_name_mapping = {
            'bookUpdates_timestamp': 'timestamp',
            'bookUpdates_aid_id': 'aid',
            'bookUpdates_aid_token_token_symbol': 'asset',
            'bookUpdates_amount': 'amount',
            'bookUpdates_amount_formatted': 'amount_formatted',
            'bookUpdates_transaction_id': 'txn',
            'bookUpdates_transaction_block': 'block',
            'bookUpdates_transaction_index': 'block_index',
            'bookUpdates_index': 'index'
        }

        if df.empty:
            df = pd.DataFrame(columns=column_name_mapping.values())
        else:
            df = df.rename(columns=column_name_mapping)
            df = df.sort_values(by=['timestamp', 'index']).reset_index(drop=True)

        return df

    def get_aid_balances(self, aid=None, asset=None, first=1000000000):
        """ a function to query the subgraph and get the most recent balances for a given aid and the tokens it trades with. 
        
        :param aid: the address of the market aid that is of interest, defaults to None
        :type aid: str, optional
        :param first: the number of balances to return, defaults to 1000000000
        :type first: int, optional
        :return: a dataframe containing the balances
        :rtype: pandas.DataFrame
        """

        aidToken = self.market_aid.AidToken
        aidToken.balance_formatted = aidToken.balance / 10 ** aidToken.token.decimals

        where = []
        if aid:
            where.append(aidToken.aid == aid.lower())
        #if asset:
        #    where.append(aidToken.token == asset)

        if where == []:
            aid_tokens = self.market_aid.Query.aidTokens(first=first)
        else:
            aid_tokens = self.market_aid.Query.aidTokens(first=first, where=where)
        
        field_paths = [
            aid_tokens.token.symbol,
            aid_tokens.balance,
            aid_tokens.balance_formatted,
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)

        # rename the columns: asset, balance, balance_formatted
        column_name_mapping = {
            'aidTokens_token_symbol': 'asset',
            'aidTokens_balance': 'balance',
            'aidTokens_balance_formatted': 'balance_formatted'
        }

        df = df.rename(columns=column_name_mapping)

        return df
    
    def get_aid_arbs(self, aid=None, asset=None, quote=None, start_time=None, end_time=None, first=1000000000):
        """ this method is intended to query arbitrage trades that occured on the market aid subgraph 
        
        :param aid: the address of the market aid that is of interest, defaults to None
        :type aid: str, optional
        :param asset: the address of the asset that is of interest, defaults to None
        :type asset: str, optional
        :param quote: the address of the quote asset that is of interest, defaults to None
        :type quote: str, optional
        :param start_time: the start time of the time range that is of interest, defaults to None
        :type start_time: int, optional
        :param end_time: the end time of the time range that is of interest, defaults to None
        :type end_time: int, optional
        :param first: the number of arbs to return, defaults to 1000000000
        :type first: int, optional
        :return: a dataframe containing the arbs
        :rtype: pandas.DataFrame
        
        """

        Arb = self.market_aid.Arb
        #Arb.amount_formatted = Arb.amount / 10 ** Arb.asset.decimals
        Arb.profit_formatted = Arb.profit / 10 ** Arb.asset.decimals

        where = []
        if aid:
            where.append(Arb.aid == aid.lower())
        if asset:
            where.append(Arb.asset == asset.lower())
        if quote:
            where.append(Arb.quote == quote.lower())
        if start_time:
            where.append(Arb.timestamp >= start_time)
        if end_time:
            where.append(Arb.timestamp <= end_time)

        if where == []:
            arbs = self.market_aid.Query.arbs(first=first)
        else:
            arbs = self.market_aid.Query.arbs(first=first, where=where)

        field_paths = [
            arbs.timestamp,
            arbs.aid.id,
            arbs.asset.symbol,
            arbs.quote.symbol,
            arbs.amount,
            arbs.profit,
            arbs.profit_formatted
        ]

        df = self.subgrounds.query_df(field_paths, pagination_strategy=ShallowStrategy)

        # if the dataframe is empty, return an empty dataframe with the correct column names
        if df.empty:
            df = pd.DataFrame(columns=['timestamp', 'aid', 'asset', 'quote', 'amount', 'profit', 'profit_formatted'])
            return df

        # rename the columns: timestamp, aid, asset, quote, amount, profit, profit_formatted
        column_name_mapping = {
            'arbs_timestamp': 'timestamp',
            'arbs_aid_id': 'aid',
            'arbs_asset_symbol': 'asset',
            'arbs_quote_symbol': 'quote',
            'arbs_amount': 'amount',
            'arbs_profit': 'profit',
            'arbs_profit_formatted': 'profit_formatted'
        }

        df = df.rename(columns=column_name_mapping)

        return df

    #def get_external_swaps()

class SuperAidData(AidData): 
    """this class acts as an extension of the AidData class and has some additional functionality that is enabled by a node connection. 
    """

    def __init__(self, w3, subgrounds, chain_id):
        """a super class of the AidData class that has additional functionality enabled by a node connection.

        :param w3: a web3 object
        :type w3: web3.main.Web3
        :param subgrounds: the subgrounds object
        :type subgrounds: subgrounds.Subgrounds
        :param chain_id: the chain id of the network that is of interest
        :type chain_id: int
        """
        super().__init__(subgrounds, chain_id)

        # set class variables
        self.w3 = w3
        self.gas = Gas(self.w3)
    
    def get_aid_txns_gas_data(self, aid=None, start_time=None, end_time=None, total_fee_eth=True, total_fee_usd=True, l2_gas_price=None, l2_gas_used=None, l1_gas_used=None, 
        l1_gas_price=None, l1_fee_scalar=None, l1_fee=None, l2_fee=None, total_fee=None, l1_fee_eth=None, l2_fee_eth=None, eth_price=None, l1_fee_usd=None, l2_fee_usd=None, first=1000000000):

        # get the transactions for the market aid contract
        txns = self.get_aid_txns(aid, start_time, end_time, first)
    
        # update the dataframe to contain the relevant gas data
        df = self.gas.txn_dataframe_update(txns, 'transactions_id', 'transactions_timestamp', total_fee_eth = total_fee_eth, total_fee_usd = total_fee_usd, l2_gas_price = l2_gas_price, l2_gas_used = l2_gas_used, l1_gas_used = l1_gas_used, l1_gas_price = l1_gas_price, l1_fee_scalar = l1_fee_scalar, l1_fee = l1_fee, l2_fee = l2_fee, total_fee = total_fee, l1_fee_eth = l1_fee_eth, l2_fee_eth = l2_fee_eth, eth_price = eth_price, l1_fee_usd = l1_fee_usd, l2_fee_usd = l2_fee_usd)

        return df 

    def get_aid_gas_spend_binned(self, aid=None, start_time=None, end_time=None, granularity=60, type = 'USD', total_fee_eth=True, total_fee_usd=True, l2_gas_price=None, l2_gas_used=None, l1_gas_used=None, 
        l1_gas_price=None, l1_fee_scalar=None, l1_fee=None, l2_fee=None, total_fee=None, l1_fee_eth=None, l2_fee_eth=None, eth_price=None, l1_fee_usd=None, l2_fee_usd=None, first=1000000000):
        """this function takes in a market aid address, start time, end time, and granularity in order to return a dictionary of the gas spend dictionary that is binned by the granularity.
        this dictionary is a key-pair mapping of the time bin (start of a minute, hour, day, etc) to the gas spend of the contract during that period.
        """

        if start_time: 
            start_time = int(start_time)
        if end_time:
            end_time = int(end_time)

        gas_spend = self.get_aid_txns_gas_data(aid, start_time, end_time, total_fee_eth, total_fee_usd, l2_gas_price, l2_gas_used, l1_gas_used, l1_gas_price, l1_fee_scalar, l1_fee, l2_fee, total_fee, l1_fee_eth, l2_fee_eth, eth_price, l1_fee_usd, l2_fee_usd, first)

        # bin the data
        gas_spend['time_bin'] = gas_spend['transactions_timestamp'].apply(lambda x: (x // granularity) * granularity)

        # group by time_bin and get the aggregate gas spend during the period represented as a dictionary with time_bin as the key and the gas spend as the value
        grouped = gas_spend.groupby(['time_bin'])

        if type == 'USD':
            gas_spend = grouped['total_fee_usd'].sum().to_dict()

        if type == 'ETH':
            gas_spend = grouped['total_fee_eth'].sum().to_dict()

        return gas_spend




import time
import requests
import asyncio
import aiohttp

class Price: 
    """Price helper class"""

    def __init__(self):
        """constructor for the price helper class
        
        :param w3: a web3 instance
        :type w3: Web3 object
        """
        self.rate_limit = 1 / 3
        self.next_delay = 0

    def get_coinbase_price(self, pair='ETH-USD', date=None): 
        """this function takes a pair and returns the current spot price for the pair from coinbase, if no date is provided
        
        :param pair: the pair to get the price for ('ETH-USD')
        :type pair: str
        :param date: the date to get the price for (YYYY-MM-DD), or converts unix timestamp to date
        :type date: str
        :return: a dictionary of the price data {'base': 'ETH', 'currency': 'USD', 'amount': '1234.56', 'time': 1673477434}
        :rtype: dict
        """

        if date and type(date) == int:
            query_date = time.strftime('%Y-%m-%d %H:%M:%S')

        if date:
            url = f'https://api.coinbase.com/v2/prices/{pair}/spot?date={query_date}'
        else:
            date = int(time.time())
            url = f'https://api.coinbase.com/v2/prices/{pair}/spot'
        
        # get the response from the api
        try:
            response = requests.get(url).json()['data']
            response['time'] = date
            response['amount'] = float(response['amount'])
        except:
            return None

        return response

    def get_coinbase_candles(self, start, end, pair='ETH-USD', granularity=60): 

        try:
            url = f"https://api.pro.coinbase.com/products/{pair}/candles?granularity={granularity}&start={start}&end={end}"
            response = requests.get(url).json()
            return response
        except:
            return None

    def get_defi_llama_price(self, network='ethereum', address='0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2', timestamp=None):
        """this function takes a network and address and returns the current spot price for the token from defi llama, if no timestamp is provided. if a timestamp is provided, it will return the price at that time.
        
        :param network: the network to get the price for ('ethereum')
        :type network: str
        :param address: the address to get the price for (mainnet weth -> '0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2')
        :type address: str
        :param timestamp: the timestamp to get the price for (unix timestamp)
        :type timestamp: int, optional
        :return: a dictionary of the price data {'base': 'ETH', 'currency': 'USD', 'amount': '1234.56', 'time': 1673477434}
        :rtype: dict
        """

        coins = f'{network}:{address}'

        if timestamp: 
            timestamp = int(timestamp)
            url = f'https://coins.llama.fi/prices/historical/{timestamp}/{coins}'
        else:
            url = f"https://coins.llama.fi/prices/current/{coins}"

        # get the response from the api
        try:
            response = requests.get(url).json()['coins'][coins]
        except: 
            return None

        # TODO: add in confidence check
        #if response['confidence'] < 0.9:
        #    return None
        #else: 
        return {'base': response['symbol'], 'currency': 'USD', 'amount': response['price'], 'time': response['timestamp']}

    async def get_prices(self, pairs, granularities, starts, ends, price_type='open'):

        async with aiohttp.ClientSession() as session:

            async def get_price(pair, granularity, start, end):

                self.next_delay += self.rate_limit
                await asyncio.sleep(self.next_delay)
                
                url = f"https://api.pro.coinbase.com/products/{pair}/candles?granularity={granularity}&start={start}&end={end}"
                try:

                    async with session.get(url) as resp: 
                        data = await resp.json()

                except aiohttp.ClientError as e:
                    raise Exception(f"Failed to retrieve data due to client error: {e}")
                except Exception as e:
                    raise Exception(f"Failed to retrieve data due to error: {e}. parameters {pair}, {granularity}, {start}, {end}")
                
                if len(data) == 0:
                    return 0

                if price_type == 'open':
                    price = data[0][3]
                elif price_type == 'high':
                    price = data[0][2]
                elif price_type == 'low':
                    price = data[0][1]
                elif price_type == 'close':
                    price = data[0][4]
                
                return price
            
            tasks = [get_price(pair, granularity, start, end) for pair, granularity, start, end in zip(pairs, granularities, starts, ends)]
            
            try:
                prices = await asyncio.gather(*tasks)
            except Exception as e:
                raise Exception(f"Failed to retrieve prices due to error: {e}")

            return prices


    def retrieve_prices(self, pairs, granularities, timestamps, price_type='open'):
        """the retrieve_prices function will take in a list of pairs, granularities, and timestamps and return a list of prices for each pair at the specified timestamp. this is done by gathering historical data from coinbase's OHLC historical data api

        :param pairs: a list of pairs to retrieve prices for, for example ['ETH-USD', 'ETH-USD', 'ETH-USD']
        :type pairs: list, str
        :param granularities: a list of granularities to retrieve prices for, for example [60, 60, 60]
        :type granularities: list, int
        :param timestamps: a list of timestamps to retrieve prices for, for example [1673931650, 1673931420, 1673931690]
        :type timestamps: list, int
        :param price_type: the price data to collect from the candle, must be one of the following: ["open", "high", "low", "close" ], defaults to "open"
        :type price_type: str, optional
        :raises Exception: an exception will be raised if the price type is not valid
        :raises Exception: an exception will be raised if the lists of pairs, granularities, and timestamps are not the same length
        :raises Exception: a generic exception will be raised if there is an error retrieving the prices
        :return: a list of prices for each pair for the specified timestamp
        :rtype: list, float
        """

        # ensure that the price type is valid
        # TODO: add in additional pricing strategies based upon the different pricing algorithms
        if price_type not in ['open', 'high', 'low', 'close']:
            raise Exception("The price type must be either 'open', 'high', 'low', or 'close'")
        
        # check that the lists are the same length, if not, raise an error
        if len(pairs) != len(granularities) or len(pairs) != len(timestamps):
            raise Exception("The lists of pairs, granularities, starts, and ends must be the same length")

        starts = [] 
        ends = []

        # based upon the granularities and timestamps, we need to determine the start and end times for each timestamp in the list
        # TODO: for today, we are going to call the api for every index of the arrays, but we will want to optimize this so that we only call the api once for each unique pair over a set granularity
        for i in range(len(timestamps)):
            period_start = (timestamps[i] // granularities[i]) * granularities[i]
            starts.append(period_start)
            ends.append(period_start)

        try: 
            try: 
                loop = asyncio.get_event_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)

            self.next_delay = 0
            prices = loop.run_until_complete(self.get_prices(pairs, granularities, starts, ends))

            return prices
        except Exception as e:
            raise Exception(f"Failed to retrieve prices due to error: {e}")

    async def get_candles_in_range(self, pair, granularity, start, end, price_type='open'):

        # add a buffer to start and end time to make sure we get all the candles we need
        # TODO: determine if there is any actual need to adding a buffer to the start and end times
        start = ((start // granularity) * granularity) # - (granularity * 10)
        end = ((end // granularity) * granularity) # + (granularity * 10)

        price_types = {'open' : 3, 'high' : 2, 'low' : 1, 'close' : 4}
        price_format = price_types[price_type]
        price_data = {}

        async with aiohttp.ClientSession() as session: 

            async def get_price(pair, granulatiry, start, end):

                self.next_delay += self.rate_limit
                await asyncio.sleep(self.next_delay)

                url = f"https://api.pro.coinbase.com/products/{pair}/candles?granularity={granularity}&start={start}&end={end}"

                try: 
                    async with session.get(url) as response:
                        data = await response.json()
                except aiohttp.ClientError as e:
                    raise Exception(f"Failed to retrieve data due to client error: {e}")
                except Exception as e:
                    raise Exception(f"Failed to retrieve data due to error: {e}. parameters {pair}, {granularity}, {start}, {end}")

                if len(data) == 0: 
                    return 0 

                for candle in data:
                    price_data[candle[0]] = candle[price_format]
                    # TODO: we will need to error handle for cases where there is no data for a given time period
                    # we will most likely want to depend upon the previous period's data to fill in the gaps
                        # there are a variety of edge cases that we will need to handle, such as when the previous period is missing data as well
                        # an example timestamp when there is no minute candle data for eth-usd is 1611929220, 1611929280, 1611929340
            
            # seperate the start and end time period into periods composed of 300 candle chunks which are defined by the granularity
            chunks = []
            chunk_size = 300 * granularity
            current_start_chunk = start
            current_end_chunk = start + chunk_size

            while current_end_chunk <= end:
                chunks.append((current_start_chunk, current_end_chunk))
                current_start_chunk = current_end_chunk
                current_end_chunk += chunk_size
            chunks.append((current_start_chunk, end))

            tasks = [get_price(pair, granularity, chunk_start, chunk_end) for chunk_start, chunk_end in chunks]

            try: 
                await asyncio.gather(*tasks)
            except Exception as e:
                raise Exception(f"Failed to retrieve data due to error: {e}")
            
            return price_data
    
    def get_price_in_range(self, start, end, pair='ETH-USD', granularity=60, price_type='open'):
        """the get_price_in_range function will return the price of a given pair at a given granularity for a given time period. it does this by collecting coinbase candlestick
        data for the given pair at the given granularity over the range of the start and end times. it then returns a dictionary of prices with the timestamp of the candle start as the key

        :param start: the start time of the time period
        :type start: int
        :param end: the end time of the time period
        :type end: int
        :param pair: the pair to retrieve the price for, defaults to 'ETH-USD'
        :type pair: str, optional
        :param granularity: the granularity of the price data, defaults to 60
        :type granularity: int, optional
        :param price_type: the type of price to return, defaults to 'open'
        :type price_type: str, optional
        :return: a dictionary of prices with the timestamp as the key and the price as the value
        :rtype: dict
        """

        # TODO: there is probably a better way to do this, but for now we need to cover the scenario where a user requests USDC-USD data as coinbase just assumes USDC is USD... i hope they are right
        if pair == 'USDC-USD':
            price_data = {}
            #TODO: figure out if there is actually any need to add a buffer to the start and end times
            start = ((start // granularity) * granularity) # - (granularity * 10)
            end = ((end // granularity) * granularity) # + (granularity * 10)
            while start <= end:
                price_data[start] = 1
                start += granularity
            return price_data
        
        try: 
            try: 
                loop = asyncio.get_event_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            
            self.next_delay = 0
            price_data = loop.run_until_complete(self.get_candles_in_range(pair, granularity, start, end, price_type))

            return price_data
        except Exception as e:
            raise Exception(f"Failed to retrieve data due to error: {e}")    

    def txn_dataframe_priced(self, txn_dataframe, timestamp_column, total_fee_usd = True, eth_price = False, l1_fee_usd = False, l2_fee_usd = False): 
        """the txn_dataframe_priced function will take in a transaction dataframe and return a dataframe with the total fee in usd, the eth price at the time of the transaction, and the l1 and l2 fees in usd

        :param txn_dataframe: a dataframe containing the transactions to price
        :type txn_dataframe: pandas.DataFrame
        :param timestamp_column: the name of the column containing the timestamp of the transaction
        :type timestamp_column: str
        :param total_fee_usd: whether or not to include the total fee in usd, defaults to True
        :type total_fee_usd: bool, optional
        :param eth_price: whether or not to include the eth price at the time of the transaction, defaults to False
        :type eth_price: bool, optional
        :param l1_fee_usd: whether or not to include the l1 fee in usd, defaults to False
        :type l1_fee_usd: bool, optional
        :param l2_fee_usd: whether or not to include the l2 fee in usd, defaults to False
        :type l2_fee_usd: bool, optional
        :return: a dataframe containing the transaction data with the total fee in usd, the eth price at the time of the transaction, and the l1 and l2 fees in usd
        :rtype: pandas.DataFrame
        """

        # get the unique timestamps from the dataframe
        timestamps = list(txn_dataframe[timestamp_column].unique())

        # create an array of "ETH-USD" strings that represent the pair, the array is the same length as the timestamps array
        pairs = ['ETH-USD'] * len(timestamps)
        granularities = [60] * len(timestamps)

        # from the timestamps and create a dictionary that maps the timestamp to the associated eth price
        eth_prices = self.retrieve_prices(pairs, granularities, timestamps)
        eth_price_data = dict(zip(timestamps, eth_prices))

        return eth_price_data

    